"""Imports of concrete (external) implementations of various NetQASM interfaces.

This module re-exports concrete types that are defined externally (e.g. in SquidASM).
The types are re-exported under generic names, such that client code does not need to
use the concrete external names.
Which of the concrete implementations is re-exported depends on global variables (most
importantly the variable indicating which simulator should be used).

For example, an application may import the `NetQASMConnection` name from
`netqasm.sdk.external`, while having set the simulator type to NETSQUID.
This results in the import resolving to an import of
`squidasm.run.multithread.sdk.NetSquidConnection`, even though the application never
had to specify this concrete name. This allows the same application code to be used
with different implementations of `NetQASMConnection`.
"""

from netqasm.runtime.settings import Simulator, get_is_using_hardware, get_simulator

simulator = get_simulator()
is_using_hardware = get_is_using_hardware()

if is_using_hardware:
    try:
        from qnodeos.sdk.connection import (
            QNodeOSConnection as NetQASMConnection,  # type: ignore
        )
        from qnodeos.sdk.socket import Socket  # type: ignore

        from netqasm.runtime.hardware import run_application  # type: ignore
    except ModuleNotFoundError:
        raise ModuleNotFoundError("to use QNodeOS , `qnodeos` needs to be installed")
elif simulator == Simulator.NETSQUID:
    try:
        from squidasm.nqasm.multithread import (
            NetSquidConnection as NetQASMConnection,  # type: ignore
        )
        from squidasm.run.multithread.simulate import (
            simulate_application,  # type: ignore
        )
        from squidasm.util.sim import get_qubit_state  # type: ignore

        from netqasm.sdk.classical_communication import (
            ThreadBroadcastChannel as BroadcastChannel,  # type: ignore
        )
        from netqasm.sdk.classical_communication import (
            ThreadSocket as Socket,  # type: ignore
        )
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            f"to use {Simulator.NETSQUID.value} as simulator, `squidasm` needs to be installed"
        )
elif simulator == Simulator.NETSQUID_SINGLE_THREAD:
    try:
        from squidasm.nqasm.singlethread.connection import (
            NetSquidConnection as NetQASMConnection,  # type: ignore
        )
        from squidasm.nqasm.singlethread.csocket import (
            NetSquidSocket as Socket,  # type: ignore
        )

        from netqasm.sdk.classical_communication import (
            ThreadBroadcastChannel as BroadcastChannel,  # type: ignore
        )

    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            f"to use {Simulator.NETSQUID.value} as simulator, `squidasm` needs to be installed"
        )
elif simulator == Simulator.SIMULAQRON:
    try:
        from simulaqron.run.run import (
            run_applications as simulate_application,  # type: ignore
        )
        from simulaqron.sdk.broadcast_channel import BroadcastChannel  # type: ignore
        from simulaqron.sdk.connection import (
            SimulaQronConnection as NetQASMConnection,  # type: ignore
        )
        from simulaqron.sdk.socket import Socket  # type: ignore
        from simulaqron.sim_util import get_qubit_state  # type: ignore
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            f"to use {Simulator.SIMULAQRON.value} as simulator, `simulaqron` needs to be installed"
        )
elif simulator == Simulator.DEBUG:
    from netqasm.runtime.debug import get_qubit_state  # type: ignore
    from netqasm.runtime.debug import run_application  # type: ignore
    from netqasm.sdk.classical_communication import (
        ThreadBroadcastChannel as BroadcastChannel,  # type: ignore
    )
    from netqasm.sdk.classical_communication import (
        ThreadSocket as Socket,  # type: ignore
    )
    from netqasm.sdk.connection import (
        DebugConnection as NetQASMConnection,  # type: ignore
    )
else:
    raise ValueError(f"Unknown simulator {simulator}")



import ctypes
from typing import Optional

from netqasm.lang import encoding
from netqasm.lang.instr import Flavour, NetQASMInstruction, VanillaFlavour
from netqasm.lang.subroutine import Subroutine

INSTR_ID = ctypes.c_uint8


class Deserializer:
    """
    Deserializes raw bytes into a Subroutine, given a Flavour.
    :class:`~.NetQASMInstructions` are immediately created from the binary encoding.

    (This is in contrast with the parsing.text module, which first converts the input
    to a :class:`~.ProtoSubroutine`, consisting of :class:`~.subroutine.ICmd` s, before transforming it into
    a :class:`~.Subroutine` containing :class:`~.NetQASMInstruction` s.)
    """

    def __init__(self, flavour: Flavour):
        self.flavour = flavour

    def _parse_metadata(self, raw):
        metadata = raw[: encoding.METADATA_BYTES]
        metadata = encoding.Metadata.from_buffer_copy(metadata)
        data = raw[encoding.METADATA_BYTES :]
        return metadata, data

    def deserialize_subroutine(self, raw: bytes) -> Subroutine:
        metadata, raw = self._parse_metadata(raw)
        if (len(raw) % encoding.COMMAND_BYTES) != 0:
            raise ValueError("Length of data not a multiple of command length")
        num_commands = int(len(raw) / encoding.COMMAND_BYTES)

        instructions = [
            self.deserialize_command(
                raw[i * encoding.COMMAND_BYTES : (i + 1) * encoding.COMMAND_BYTES]
            )
            for i in range(num_commands)
        ]

        return Subroutine(
            netqasm_version=tuple(metadata.netqasm_version),  # type: ignore
            app_id=metadata.app_id,  # type: ignore
            instructions=instructions,
        )

    def deserialize_command(self, raw: bytes) -> NetQASMInstruction:
        # peek next byte to check instruction type
        id = INSTR_ID.from_buffer_copy(raw[:1]).value

        # use the flavour to check which NetQASMInstruction class should be created
        instr_cls = self.flavour.get_instr_by_id(id)
        instr: NetQASMInstruction = instr_cls.deserialize_from(raw)
        return instr


def deserialize(data: bytes, flavour: Optional[Flavour] = None) -> Subroutine:
    """
    Convert a binary encoding into a Subroutine object.
    The Vanilla flavour is used by default.
    """
    if flavour is None:
        flavour = VanillaFlavour()

    return Deserializer(flavour).deserialize_subroutine(data)



from typing import Dict, Set

import numpy as np
import pytest

from netqasm.backend.messages import deserialize_host_msg as deserialize_message
from netqasm.lang.instr import core
from netqasm.lang.instr.flavour import NVFlavour, VanillaFlavour
from netqasm.lang.operand import Register
from netqasm.lang.parsing import deserialize as deserialize_subroutine
from netqasm.lang.parsing.text import parse_text_subroutine
from netqasm.lang.subroutine import Subroutine
from netqasm.logging.glob import set_log_level
from netqasm.sdk.connection import DebugConnection
from netqasm.sdk.qubit import Qubit
from netqasm.sdk.transpile import NVSubroutineTranspiler


def pad_single_matrix(m: np.ndarray, index: int, total: int) -> np.ndarray:
    """Create matrix for `total` qubits where `m` is applied on qubit `index`
    and identities on the other qubits"""
    matrix = np.eye(1)
    for i in range(total):
        if i == index:
            matrix = np.kron(matrix, m)
        else:
            matrix = np.kron(matrix, np.eye(2))
    return matrix


def pad_controlled_single_matrix(
    m: np.ndarray, ctrl_index, target_index, total
) -> np.ndarray:
    """Create matrix for `total` qubits where `m` is applied on qubit `target_index`
    if qubit in `ctrl_index` is 1 (and identities on the other qubits).
    E.g. pad_controlled_single_matrix(X, 3, 1, 3) returns a 4-qubit matrix representing a CNOT
    from qubit 3 to 1, while qubits 0 and 2 are not affected (identities).
    """
    matrix0 = np.eye(1)
    for i in range(total):
        if i == ctrl_index:
            matrix0 = np.kron(matrix0, np.array([[1, 0], [0, 0]]))
        else:
            matrix0 = np.kron(matrix0, np.eye(2))

    matrix1 = np.eye(1)
    for i in range(total):
        if i == ctrl_index:
            matrix1 = np.kron(matrix1, np.array([[0, 0], [0, 1]]))
        elif i == target_index:
            matrix1 = np.kron(matrix1, m)
        else:
            matrix1 = np.kron(matrix1, np.eye(2))

    return matrix0 + matrix1


class SubroutineMatrix:
    """Matrix representing the quantum operations in a subroutine"""

    def __init__(self, virt_ids: Set[int]):
        self._virt_ids = virt_ids
        self._matrix_indices: Dict[
            int, int
        ] = dict()  # map of virt ID to index in matrix

        self._matrix = np.eye(1)
        for i, id in enumerate(virt_ids):
            self._matrix_indices[id] = i
            self._matrix = np.kron(self._matrix, np.eye(2))

    def apply_single_qubit_instr(self, instr_matrix: np.ndarray, virt_id: int):
        m = pad_single_matrix(
            m=instr_matrix,
            index=self._matrix_indices[virt_id],
            total=len(self._matrix_indices.keys()),
        )

        self._matrix = self._matrix @ m

    def apply_two_qubit_instr(
        self, instr_matrix: np.ndarray, virt_id0: int, virt_id1: int
    ):
        m = pad_controlled_single_matrix(
            m=instr_matrix,
            ctrl_index=self._matrix_indices[virt_id0],
            target_index=self._matrix_indices[virt_id1],
            total=len(self._matrix_indices.keys()),
        )

        self._matrix = self._matrix @ m

    @property
    def matrix(self):
        return self._matrix


def _subroutine_as_matrix(subroutine: Subroutine) -> np.ndarray:
    """Try to write quantum instructions in subroutine as one matrix.
    This function relies on the fact that the subroutine has only "set"
    instructions for Q registers, and should only be used for subroutines
    specifically made for testing the NV transpiler.
    """
    qreg_values: Dict[Register, int] = dict()
    virt_ids: Set[int] = set()

    # first pass: collect virt IDs used
    for instr in subroutine.instructions:
        if isinstance(instr, core.SetInstruction):
            virt_ids.add(instr.imm.value)

    sub_matrix = SubroutineMatrix(virt_ids)

    # second pass: keep track of qreg values and construct matrix
    for instr in subroutine.instructions:
        if isinstance(instr, core.SetInstruction):
            qreg_values[instr.reg] = instr.imm.value

        if isinstance(instr, core.SingleQubitInstruction):
            sub_matrix.apply_single_qubit_instr(
                instr.to_matrix(), qreg_values[instr.reg]
            )
        elif isinstance(instr, core.RotationInstruction):
            sub_matrix.apply_single_qubit_instr(
                instr.to_matrix(), qreg_values[instr.reg]
            )
        elif isinstance(instr, core.TwoQubitInstruction):
            sub_matrix.apply_two_qubit_instr(
                instr_matrix=instr.to_matrix_target_only(),
                virt_id0=qreg_values[instr.reg0],
                virt_id1=qreg_values[instr.reg1],
            )

    return sub_matrix.matrix  # type: ignore


@pytest.mark.parametrize(
    "text_subroutine",
    [
        (
            """
        # NETQASM 0.0
        # APPID 0
        set Q0 0
        z Q0
        """
        ),
        (
            """
        # NETQASM 0.0
        # APPID 0
        set Q0 0
        set Q1 1
        rot_x Q0 3 1
        """
        ),
        (
            """
        # NETQASM 0.0
        # APPID 0
        set Q0 0
        set Q1 1
        cnot Q1 Q0
        """
        ),
        (
            """
        # NETQASM 0.0
        # APPID 0
        set Q0 0
        set Q1 1
        cphase Q1 Q0
        """
        ),
        (
            """
        # NETQASM 0.0
        # APPID 0
        set Q0 0
        set Q1 1
        set Q2 2
        cphase Q1 Q2
        h Q1
        cnot Q2 Q0
        """
        ),
    ],
)
def test_mapping(text_subroutine: str):
    """
    Test whether the NV transpiler correctly maps gates by comparing the matrices
    representing these gates. For this it's sufficient to have simple (incomplete, e.g. no alloc) subroutines.
    """
    vanilla_subroutine = parse_text_subroutine(text_subroutine)
    vanilla_matrix = _subroutine_as_matrix(vanilla_subroutine)
    print(f"vanilla: {vanilla_matrix}")

    transpiled_subroutine = NVSubroutineTranspiler(vanilla_subroutine).transpile()
    transpiled_matrix = _subroutine_as_matrix(transpiled_subroutine)
    print(f"transpiled: {np.round(transpiled_matrix, 2)}")
    print(transpiled_subroutine)

    assert True  # TODO: test mapping of controlled-rotation gates
    # assert are_matrices_equal(vanilla_matrix, transpiled_matrix)


def test_transpiling_nv():
    text_subroutine = """
# NETQASM 0.0
# APPID 0
set Q0 0
qalloc Q0
init Q0
x Q0
y Q0
z Q0
h Q0
k Q0
s Q0
t Q0
rot_x Q0 1 2
rot_y Q0 1 2
rot_z Q0 1 2
"""
    original_subroutine = parse_text_subroutine(text_subroutine)
    print(f"before transpiling: {original_subroutine}")
    subroutine = NVSubroutineTranspiler(original_subroutine).transpile()
    print(f"after transpiling: {subroutine}")

    for instr in subroutine.instructions:
        assert instr.__class__ not in VanillaFlavour().instrs


@pytest.mark.parametrize(
    "subroutine_str",
    [
        (
            """
        # NETQASM 0.0
        # APPID 0
        set Q0 0
        qalloc Q0
        init Q0
        set Q1 1
        qalloc Q1
        init Q1
        cnot Q0 Q1
        """
        ),
        (
            """
        # NETQASM 0.0
        # APPID 0
        set Q0 0
        qalloc Q0
        init Q0
        set Q1 1
        qalloc Q1
        init Q1
        cphase Q1 Q0
        """
        ),
    ],
)
def test_transpiling_nv_text(subroutine_str):
    original = parse_text_subroutine(subroutine_str)
    print(f"before transpiling: {original}")
    transpiled = NVSubroutineTranspiler(original).transpile()
    print(f"after transpiling: {transpiled}")

    for instr in transpiled.instructions:
        assert instr.__class__ not in VanillaFlavour().instrs


def test_transpiling_nv_using_sdk():
    set_log_level("DEBUG")
    with DebugConnection("Alice", compiler=NVSubroutineTranspiler) as alice:
        q = Qubit(alice)
        q.X()
        q.Y()
        q.Z()
        q.H()
        q.K()
        q.S()
        q.T()
        q.rot_X(n=1, d=2)
        q.rot_Y(n=1, d=2)
        q.rot_Z(n=1, d=2)

    assert len(alice.storage) == 4
    raw_subroutine = deserialize_message(raw=alice.storage[1]).subroutine
    subroutine = deserialize_subroutine(raw_subroutine, flavour=NVFlavour())

    # NOTE this does not test much anymore since we need to state which flavour we
    # are using to be able to deserialize
    for instr in subroutine.instructions:
        assert instr.__class__ not in VanillaFlavour().instrs


if __name__ == "__main__":
    test_transpiling_nv()
    test_transpiling_nv_using_sdk()



from netqasm.logging.glob import get_netqasm_logger
from netqasm.sdk import EPRSocket
from netqasm.sdk.external import NetQASMConnection
from netqasm.sdk.toolbox.measurements import parity_meas

logger = get_netqasm_logger()


def _get_default_strategy():
    return [
        ["XI", "XX", "IX"],  # row 0
        ["-XZ", "YY", "-ZX"],  # row 1
        ["IZ", "ZZ", "ZI"],  # row 2
    ]


def main(app_config=None, row=0, strategy=None):

    if strategy is None:
        strategy = _get_default_strategy()
    if row >= len(strategy):
        raise ValueError(f"Not a row in the square {row}")

    # Create a EPR socket for entanglement generation
    epr_socket = EPRSocket("player2")

    # Initialize the connection
    player1 = NetQASMConnection(
        app_name=app_config.app_name,
        log_config=app_config.log_config,
        epr_sockets=[epr_socket],
    )
    with player1:
        # Create EPR pairs
        q1 = epr_socket.create_keep()[0]
        q2 = epr_socket.create_keep()[0]

        # TODO put in single subroutine?
        player1.flush()

        # Make sure we order the qubits consistently with player2
        # Get entanglement IDs
        q1_ID = q1.entanglement_info.sequence_number
        q2_ID = q2.entanglement_info.sequence_number

        if int(q1_ID) < int(q2_ID):
            qa = q1
            qc = q2
        else:
            qa = q2
            qc = q1

        # Perform the three measurements
        m0, m1, m2 = (parity_meas([qa, qc], strategy[row][i]) for i in range(3))

    to_print = "\n\n"
    to_print += "==========================\n"
    to_print += "App player1: row is:\n"
    for _ in range(row):
        to_print += "(___)\n"
    to_print += f"({m0}{m1}{m2})\n"
    for _ in range(2 - row):
        to_print += "(___)\n"
    to_print += "==========================\n"
    to_print += "\n\n"
    logger.info(to_print)

    return {
        "row": [int(m0), int(m1), int(m2)],
    }


if __name__ == "__main__":
    main()



from shared.myfuncs import custom_measure, custom_send

from netqasm.logging.glob import get_netqasm_logger
from netqasm.sdk import Qubit
from netqasm.sdk.external import NetQASMConnection, Socket

logger = get_netqasm_logger()


def main(app_config=None):
    socket = Socket("alice", "bob", log_config=app_config.log_config)

    # Initialize the connection to the backend
    alice = NetQASMConnection(
        app_name=app_config.app_name, log_config=app_config.log_config
    )

    with alice:
        q1 = Qubit(alice)
        q1.measure()
        q2 = Qubit(alice)
        custom_measure(q2)

    socket.send("hello from main()")
    custom_send(socket)


if __name__ == "__main__":
    main()



from abc import abstractmethod
from dataclasses import dataclass
from typing import List, Union

import numpy as np

from netqasm.lang.operand import Immediate, Operand, Register, Template

from . import base

# Explicit core NetQASM instructions.


@dataclass  # type: ignore
class SingleQubitInstruction(base.RegInstruction):
    @property
    def qreg(self):
        return self.reg

    @qreg.setter
    def qreg(self, new_val: Register):
        self.reg = new_val

    @abstractmethod
    def to_matrix(self) -> np.ndarray:
        pass


@dataclass  # type: ignore
class TwoQubitInstruction(base.RegRegInstruction):
    @property
    def qreg0(self):
        return self.reg0

    @qreg0.setter
    def qreg0(self, new_val: Register):
        self.reg0 = new_val

    @property
    def qreg1(self):
        return self.reg1

    @qreg1.setter
    def qreg1(self, new_val: Register):
        self.reg1 = new_val

    @abstractmethod
    def to_matrix(self):
        pass

    @abstractmethod
    def to_matrix_target_only(self):
        pass


@dataclass  # type: ignore
class RotationInstruction(base.RegImmImmInstruction):
    @property
    def qreg(self):
        return self.reg

    @qreg.setter
    def qreg(self, new_val: Register):
        self.reg = new_val

    @property
    def angle_num(self):
        return self.imm0

    @angle_num.setter
    def angle_num(self, new_val: Immediate):
        self.imm0 = new_val

    @property
    def angle_denom(self):
        return self.imm1

    @angle_denom.setter
    def angle_denom(self, new_val: Immediate):
        self.imm1 = new_val

    @abstractmethod
    def to_matrix(self):
        pass

    @classmethod
    def from_operands(cls, operands: List[Union[Operand, int]]):
        assert len(operands) == 3
        reg, imm0, imm1 = operands
        assert isinstance(reg, Register)

        if isinstance(imm0, int):
            imm0 = Immediate(value=imm0)
        elif isinstance(imm0, Immediate):
            pass
        else:
            assert isinstance(imm0, Template) or isinstance(imm0, Register)
        if isinstance(imm1, int):
            imm1 = Immediate(value=imm1)
        elif isinstance(imm1, Immediate):
            pass
        else:
            assert isinstance(imm1, Template) or isinstance(imm1, Register)
        # We allow imm0, imm1 to be Templates OR registers
        return cls(reg=reg, imm0=imm0, imm1=imm1)  # type: ignore


@dataclass  # type: ignore
class ControlledRotationInstruction(base.RegRegImmImmInstruction):
    @property
    def qreg0(self):
        return self.reg0

    @qreg0.setter
    def qreg0(self, new_val: Register):
        self.reg0 = new_val

    @property
    def qreg1(self):
        return self.reg1

    @qreg1.setter
    def qreg1(self, new_val: Register):
        self.reg1 = new_val

    @property
    def angle_num(self):
        return self.imm0

    @angle_num.setter
    def angle_num(self, new_val: Immediate):
        self.imm0 = new_val

    @property
    def angle_denom(self):
        return self.imm1

    @angle_denom.setter
    def angle_denom(self, new_val: Immediate):
        self.imm1 = new_val

    @abstractmethod
    def to_matrix(self):
        pass


@dataclass
class ClassicalOpInstruction(base.RegRegRegInstruction):
    def writes_to(self) -> List[Register]:
        return [self.regout]

    @property
    def regout(self):
        return self.reg0

    @regout.setter
    def regout(self, new_val: Register):
        self.reg0 = new_val

    @property
    def regin0(self):
        return self.reg1

    @regin0.setter
    def regin0(self, new_val: Register):
        self.reg1 = new_val

    @property
    def regin1(self):
        return self.reg2

    @regin1.setter
    def regin1(self, new_val: Register):
        self.reg2 = new_val


@dataclass
class ClassicalOpModInstruction(base.RegRegRegRegInstruction):
    def writes_to(self) -> List[Register]:
        return [self.regout]

    @property
    def regout(self):
        return self.reg0

    @regout.setter
    def regout(self, new_val: Register):
        self.reg0 = new_val

    @property
    def regin0(self):
        return self.reg1

    @regin0.setter
    def regin0(self, new_val: Register):
        self.reg1 = new_val

    @property
    def regin1(self):
        return self.reg2

    @regin1.setter
    def regin1(self, new_val: Register):
        self.reg2 = new_val

    @property
    def regmod(self):
        return self.reg3

    @regmod.setter
    def regmod(self, new_val: Register):
        self.reg3 = new_val


@dataclass
class QAllocInstruction(base.RegInstruction):
    id: int = 1
    mnemonic: str = "qalloc"

    @property
    def qreg(self):
        return self.reg

    @qreg.setter
    def qreg(self, new_val: Register):
        self.reg = new_val


@dataclass
class InitInstruction(base.RegInstruction):
    id: int = 2
    mnemonic: str = "init"

    @property
    def qreg(self):
        return self.reg

    @qreg.setter
    def qreg(self, new_val: Register):
        self.reg = new_val


@dataclass
class ArrayInstruction(base.RegAddrInstruction):
    id: int = 3
    mnemonic: str = "array"

    @property
    def size(self):
        return self.reg

    @size.setter
    def size(self, new_val):
        self.reg = new_val


@dataclass
class SetInstruction(base.RegImmInstruction):
    id: int = 4
    mnemonic: str = "set"

    def writes_to(self) -> List[Register]:
        return [self.reg]

    @classmethod
    def from_operands(cls, operands: List[Union[Operand, int]]):
        assert len(operands) == 2
        reg, imm = operands
        assert isinstance(reg, Register)

        if isinstance(imm, int):
            imm = Immediate(value=imm)
        elif isinstance(imm, Immediate):
            pass
        else:
            assert isinstance(imm, Template)

        return cls(reg=reg, imm=imm)  # type: ignore


@dataclass
class StoreInstruction(base.RegEntryInstruction):
    id: int = 5
    mnemonic: str = "store"


@dataclass
class LoadInstruction(base.RegEntryInstruction):
    id: int = 6
    mnemonic: str = "load"

    def writes_to(self) -> List[Register]:
        return [self.reg]


@dataclass
class UndefInstruction(base.ArrayEntryInstruction):
    id: int = 7
    mnemonic: str = "undef"


@dataclass
class LeaInstruction(base.RegAddrInstruction):
    id: int = 8
    mnemonic: str = "lea"

    def writes_to(self) -> List[Register]:
        return [self.reg]


@dataclass
class JmpInstruction(base.ImmInstruction):
    id: int = 9
    mnemonic: str = "jmp"

    @property
    def line(self):
        return self.imm

    @line.setter
    def line(self, new_val):
        self.imm = new_val


@dataclass  # type: ignore
class BranchUnaryInstruction(base.RegImmInstruction):
    """
    Represents an instruction to branch to a certain line, depending on a
    unary expression.
    """

    @property
    def line(self):
        return self.imm

    @line.setter
    def line(self, new_val):
        self.imm = new_val

    @abstractmethod
    def check_condition(self, a) -> bool:
        raise RuntimeError(
            "check_condition called on the base BranchUnaryInstruction class"
        )


@dataclass
class BezInstruction(BranchUnaryInstruction):
    id: int = 10
    mnemonic: str = "bez"

    def check_condition(self, a: int) -> bool:
        return a == 0


@dataclass
class BnzInstruction(BranchUnaryInstruction):
    id: int = 11
    mnemonic: str = "bnz"

    def check_condition(self, a: int) -> bool:
        return a != 0


@dataclass  # type: ignore
class BranchBinaryInstruction(base.RegRegImmInstruction):
    """
    Represents an instruction to branch to a certain line, depending on a
    binary expression.
    """

    @property
    def line(self):
        return self.imm

    @line.setter
    def line(self, new_val):
        self.imm = new_val

    @abstractmethod
    def check_condition(self, a, b) -> bool:
        raise RuntimeError(
            "check_condition called on the base BranchBinaryInstruction class"
        )


@dataclass
class BeqInstruction(BranchBinaryInstruction):
    id: int = 12
    mnemonic: str = "beq"

    def check_condition(self, a: int, b: int) -> bool:
        return a == b


@dataclass
class BneInstruction(BranchBinaryInstruction):
    id: int = 13
    mnemonic: str = "bne"

    def check_condition(self, a: int, b: int) -> bool:
        return a != b


@dataclass
class BltInstruction(BranchBinaryInstruction):
    id: int = 14
    mnemonic: str = "blt"

    def check_condition(self, a: int, b: int) -> bool:
        return a < b


@dataclass
class BgeInstruction(BranchBinaryInstruction):
    id: int = 15
    mnemonic: str = "bge"

    def check_condition(self, a: int, b: int) -> bool:
        return a >= b


@dataclass
class AddInstruction(ClassicalOpInstruction):
    id: int = 16
    mnemonic: str = "add"


@dataclass
class SubInstruction(ClassicalOpInstruction):
    id: int = 17
    mnemonic: str = "sub"


@dataclass
class AddmInstruction(ClassicalOpModInstruction):
    id: int = 18
    mnemonic: str = "addm"


@dataclass
class SubmInstruction(ClassicalOpModInstruction):
    id: int = 19
    mnemonic: str = "subm"


@dataclass
class MulInstruction(ClassicalOpInstruction):
    id: int = 200
    mnemonic: str = "mul"


@dataclass
class DivInstruction(ClassicalOpInstruction):
    id: int = 201
    mnemonic: str = "div"


@dataclass
class RemInstruction(ClassicalOpInstruction):
    id: int = 202
    mnemonic: str = "rem"


@dataclass
class MeasInstruction(base.RegRegInstruction):
    id: int = 32
    mnemonic: str = "meas"

    def writes_to(self) -> List[Register]:
        return [self.creg]

    @property
    def qreg(self):
        return self.reg0

    @qreg.setter
    def qreg(self, new_val: Register):
        self.reg0 = new_val

    @property
    def creg(self):
        return self.reg1

    @creg.setter
    def creg(self, new_val: Register):
        self.reg1 = new_val


@dataclass
class MeasBasisInstruction(base.RegRegImm4Instruction):
    id: int = 41
    mnemonic: str = "meas_basis"

    def writes_to(self) -> List[Register]:
        return [self.creg]

    @property
    def qreg(self):
        return self.reg0

    @qreg.setter
    def qreg(self, new_val: Register):
        self.reg0 = new_val

    @property
    def creg(self):
        return self.reg1

    @creg.setter
    def creg(self, new_val: Register):
        self.reg1 = new_val

    @property
    def angle_num_x1(self):
        return self.imm0

    @angle_num_x1.setter
    def angle_num_x1(self, new_val: Immediate):
        self.imm0 = new_val

    @property
    def angle_num_y(self):
        return self.imm1

    @angle_num_y.setter
    def angle_num_y(self, new_val: Immediate):
        self.imm1 = new_val

    @property
    def angle_num_x2(self):
        return self.imm2

    @angle_num_x2.setter
    def angle_num_x2(self, new_val: Immediate):
        self.imm2 = new_val

    @property
    def angle_denom(self):
        return self.imm3

    @angle_denom.setter
    def angle_denom(self, new_val: Immediate):
        self.imm3 = new_val


@dataclass
class CreateEPRInstruction(base.Reg5Instruction):
    id: int = 33
    mnemonic: str = "create_epr"

    @property
    def remote_node_id(self):
        return self.reg0

    @remote_node_id.setter
    def remote_node_id(self, new_val: Register):
        self.reg0 = new_val

    @property
    def epr_socket_id(self):
        return self.reg1

    @epr_socket_id.setter
    def epr_socket_id(self, new_val: Register):
        self.reg1 = new_val

    @property
    def qubit_addr_array(self):
        return self.reg2

    @qubit_addr_array.setter
    def qubit_addr_array(self, new_val: Register):
        self.reg2 = new_val

    @property
    def arg_array(self):
        return self.reg3

    @arg_array.setter
    def arg_array(self, new_val: Register):
        self.reg3 = new_val

    @property
    def ent_results_array(self):
        return self.reg4

    @ent_results_array.setter
    def ent_results_array(self, new_val: Register):
        self.reg4 = new_val


@dataclass
class RecvEPRInstruction(base.RegRegRegRegInstruction):
    id: int = 34
    mnemonic: str = "recv_epr"

    @property
    def remote_node_id(self):
        return self.reg0

    @remote_node_id.setter
    def remote_node_id(self, new_val: Register):
        self.reg0 = new_val

    @property
    def epr_socket_id(self):
        return self.reg1

    @epr_socket_id.setter
    def epr_socket_id(self, new_val: Register):
        self.reg1 = new_val

    @property
    def qubit_addr_array(self):
        return self.reg2

    @qubit_addr_array.setter
    def qubit_addr_array(self, new_val: Register):
        self.reg2 = new_val

    @property
    def ent_results_array(self):
        return self.reg3

    @ent_results_array.setter
    def ent_results_array(self, new_val: Register):
        self.reg3 = new_val


@dataclass
class WaitAllInstruction(base.ArraySliceInstruction):
    id: int = 35
    mnemonic: str = "wait_all"


@dataclass
class WaitAnyInstruction(base.ArraySliceInstruction):
    id: int = 36
    mnemonic: str = "wait_any"


@dataclass
class WaitSingleInstruction(base.ArrayEntryInstruction):
    id: int = 37
    mnemonic: str = "wait_single"


@dataclass
class QFreeInstruction(base.RegInstruction):
    id: int = 38
    mnemonic: str = "qfree"

    @property
    def qreg(self):
        return self.reg

    @qreg.setter
    def qreg(self, new_val: Register):
        self.reg = new_val


@dataclass
class RetRegInstruction(base.RegInstruction):
    id: int = 39
    mnemonic: str = "ret_reg"


@dataclass
class RetArrInstruction(base.AddrInstruction):
    id: int = 40
    mnemonic: str = "ret_arr"


@dataclass
class BreakpointInstruction(base.ImmImmInstruction):
    id: int = 100
    mnemonic: str = "breakpoint"

    @property
    def action(self):
        return self.imm0

    @action.setter
    def action(self, new_val):
        self.imm0 = new_val

    @property
    def role(self):
        return self.imm1

    @role.setter
    def role(self, new_val):
        self.imm1 = new_val



from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, List

from netqasm.util.yaml import load_yaml


class QuantumHardware(Enum):
    Generic = "Generic"
    NV = "NV"
    TrappedIon = "TrappedIon"


class NoiseType(Enum):
    NoNoise = "NoNoise"
    Depolarise = "Depolarise"
    DiscreteDepolarise = "DiscreteDepolarise"
    Bitflip = "Bitflip"


@dataclass
class Qubit:
    id: int
    t1: float
    t2: float


@dataclass
class Node:
    name: str
    hardware: QuantumHardware
    qubits: List[Qubit]
    gate_fidelity: float = 1.0


@dataclass
class Link:
    name: str
    node_name1: str
    node_name2: str
    noise_type: NoiseType
    fidelity: float


# Network configuration.
@dataclass
class NetworkConfig:
    nodes: List[Node]
    links: List[Link]


_DEFAULT_NUM_QUBITS = 5


def default_network_config(
    node_names: List[str], hardware: QuantumHardware = QuantumHardware.Generic
) -> NetworkConfig:
    """Create a config for a fully connected network of nodes with the given names"""
    nodes = []
    links = []
    for name in node_names:
        qubits = [Qubit(id=i, t1=0, t2=0) for i in range(_DEFAULT_NUM_QUBITS)]
        node = Node(name=name, hardware=hardware, qubits=qubits, gate_fidelity=1)
        nodes += [node]

        for other_name in node_names:
            if other_name == name:
                continue
            link = Link(
                name=f"link_{name}_{other_name}",
                node_name1=name,
                node_name2=other_name,
                noise_type=NoiseType.NoNoise,
                fidelity=1,
            )
            links += [link]

    return NetworkConfig(nodes, links)


# Build a NetworkConfig object from a dict that is created by reading a yaml file.
def parse_network_config(cfg) -> NetworkConfig:
    try:
        node_cfgs = cfg["nodes"]
        link_cfgs = cfg["links"]

        nodes = []
        for node_cfg in node_cfgs:
            qubit_cfgs = node_cfg["qubits"]
            qubits = []
            for qubit_cfg in qubit_cfgs:
                qubit = Qubit(
                    id=qubit_cfg["id"],
                    t1=qubit_cfg["t1"],
                    t2=qubit_cfg["t2"],
                )
                qubits += [qubit]
            hardware = node_cfg.get("hardware", QuantumHardware.Generic)

            node = Node(
                name=node_cfg["name"],
                hardware=hardware,
                qubits=qubits,
                gate_fidelity=node_cfg["gate_fidelity"],
            )
            nodes += [node]

        links = []
        for link_cfg in link_cfgs:
            link = Link(
                name=link_cfg["name"],
                node_name1=link_cfg["node_name1"],
                node_name2=link_cfg["node_name2"],
                noise_type=link_cfg["noise_type"],
                fidelity=link_cfg["fidelity"],
            )
            links += [link]
    except KeyError as e:
        raise ValueError(f"Invalid network configuration: key not found: {e}")

    return NetworkConfig(nodes, links)


# Role-node configuration.
# Keys are role names, values are node names.
RolesConfig = Dict[str, str]


# App input per role.
# Keys are names of variables that are passed to the role's `main` function.
AppInput = Dict[str, Any]


def network_cfg_from_file(network_config_file: str = None) -> NetworkConfig:
    yaml_dict = load_yaml(network_config_file)
    network_cfg = parse_network_config(yaml_dict)
    return network_cfg



from netqasm.logging.glob import get_netqasm_logger
from netqasm.runtime.application import default_app_instance
from netqasm.sdk import Qubit
from netqasm.sdk.external import NetQASMConnection, simulate_application

logger = get_netqasm_logger()


def run_alice():
    logger.debug("Starting Alice thread")
    with NetQASMConnection("Alice") as alice:
        q1 = Qubit(alice)
        q2 = Qubit(alice)
        q1.H()
        q2.X()
        q1.X()
        q2.H()
    assert len(alice.active_qubits) == 0
    logger.debug("End Alice thread")


def run_bob():
    logger.debug("Starting Bob thread")
    with NetQASMConnection("Bob") as bob:
        q1 = Qubit(bob)
        q2 = Qubit(bob)
        q1.H()
        q2.X()
        q1.X()
        q2.H()
    assert len(bob.active_qubits) == 0
    logger.debug("End Bob thread")


def test_two_nodes():
    app_instance = default_app_instance(
        [
            ("Alice", run_alice),
            ("Bob", run_bob),
        ]
    )
    simulate_application(app_instance, use_app_config=False, enable_logging=False)



from shared.myfuncs import custom_measure, custom_recv

from netqasm.logging.glob import get_netqasm_logger
from netqasm.sdk import Qubit
from netqasm.sdk.external import NetQASMConnection, Socket

logger = get_netqasm_logger()


def main(app_config=None):
    socket = Socket("bob", "alice", log_config=app_config.log_config)

    # Initialize the connection to the backend
    bob = NetQASMConnection(
        app_name=app_config.app_name, log_config=app_config.log_config
    )
    with bob:
        q = Qubit(bob)
        custom_measure(q)

    socket.recv()
    custom_recv(socket)


if __name__ == "__main__":
    main()



from netqasm.logging.glob import get_netqasm_logger
from netqasm.sdk import EPRSocket
from netqasm.sdk.external import NetQASMConnection

logger = get_netqasm_logger()


def main(app_config=None):
    epr_socket_alice = EPRSocket(
        remote_app_name="alice", epr_socket_id=0, remote_epr_socket_id=0
    )
    epr_socket_charlie = EPRSocket(
        remote_app_name="charlie", epr_socket_id=1, remote_epr_socket_id=1
    )

    bob = NetQASMConnection(
        app_name=app_config.app_name,
        log_config=app_config.log_config,
        epr_sockets=[epr_socket_alice, epr_socket_charlie],
    )
    with bob:
        epr_alice = epr_socket_alice.recv_keep()[0]
        m_alice = epr_alice.measure()

        bob.flush()

        epr_charlie = epr_socket_charlie.create_keep()[0]
        m_charlie = epr_charlie.measure()

    logger.info(f"bob:      m_alice:  {m_alice}")
    logger.info(f"bob:      m_charlie:{m_charlie}")


if __name__ == "__main__":
    main()



"""Interface for classical communication between Hosts.

This module contains the `Socket` class which is a base for representing classical
communication (sending and receiving classical messages) between Hosts.
"""
from __future__ import annotations

import abc
from typing import TYPE_CHECKING, Optional

from netqasm.sdk.classical_communication.message import StructuredMessage

if TYPE_CHECKING:
    from netqasm.sdk import config


class Socket(abc.ABC):
    """Base class for classical sockets.

    Classical communication is modelled by sockets, which are also widely used
    in purely classical applications involving communication.

    If a node wants to communicate arbitrary classical messages with another node, this
    communication must be done between the Hosts of these nodes. Both Hosts should
    instantiate a Socket object with the other Host as 'remote'. Upon creation, the
    Socket objects try to connect to each other. Only after this has succeeded, the
    sockets can be used.

    The main methods of a Socket are `send` and `recv`, which are used to send a
    message to the remote Host, and wait to receive a message from the other Host,
    respectively. Messages are str (string) objects. To send a number, convert it
    to a string before sending, and convert it back after receiving.

    There are some variations on the `send` and `recv` methods which may be useful
    in specific scenarios. See their own documentation for their use.

    NOTE: At the moment, Sockets are not part of the compilation process yet. Therefore,
    they don't need to be part of a Connection, and operations on Sockets do not need
    to be flushed before they are executed (they are executed immediately).
    This also means that e.g. a `recv` operation, which is blocking by default,
    acutally blocks the whole application script. **So, if any quantum operations
    should be executed before such a `recv` statement, make sure that these operations
    are flushed before blocking on `recv`**.

    Implementations (subclasses) of Sockets may be quite different, depending on the
    runtime environment. A physical setup (with real hardware) may implement this as
    TCP sockets. A simulator might use inter-thread communication (see e.g.
    `ThreadSocket`), or another custom object type.
    """

    def __init__(
        self,
        app_name: str,
        remote_app_name: str,
        socket_id: int = 0,
        timeout: Optional[float] = None,
        use_callbacks: bool = False,
        log_config: Optional[config.LogConfig] = None,
    ):
        """Socket constructor.

        :param app_name: application/Host name of this socket's owner
        :param remote_app_name: application/Host name of this socket's remote
        :param socket_id: local ID to use for this socket
        :param timeout: maximum amount of real time to try to connect to the remote
            socket
        :param use_callbacks: whether to call the `recv_callback` method upon receiving
            a message
        :param log_config: logging configuration for this socket
        """
        pass

    @abc.abstractmethod
    def send(self, msg: str) -> None:
        """Send a message to the remote node."""
        pass

    @abc.abstractmethod
    def recv(
        self,
        block: bool = True,
        timeout: Optional[float] = None,
        maxsize: Optional[int] = None,
    ) -> str:
        """Receive a message from the remote node."""
        pass

    def send_structured(self, msg: StructuredMessage) -> None:
        """Sends a structured message (with header and payload) to the remote node."""
        raise NotImplementedError

    def recv_structured(
        self,
        block: bool = True,
        timeout: Optional[float] = None,
        maxsize: Optional[int] = None,
    ) -> StructuredMessage:
        """Receive a message (with header and payload) from the remote node."""
        raise NotImplementedError

    def send_silent(self, msg: str) -> None:
        """Sends a message without logging"""
        raise NotImplementedError

    def recv_silent(
        self, block: bool = True, timeout: Optional[float] = None, maxsize: int = None
    ) -> str:
        """Receive a message without logging"""
        raise NotImplementedError

    def recv_callback(self, msg: str) -> None:
        """This method gets called when a message is received.

        Subclass to define behaviour.

        NOTE: This only happens if `self.use_callbacks` is set to `True`.
        """
        pass

    def conn_lost_callback(self) -> None:
        """This method gets called when the connection is lost.

        Subclass to define behaviour.

        NOTE: This only happens if `self.use_callbacks` is set to `True`.
        """
        pass



from netqasm.lang.operand import Template
from netqasm.logging.glob import set_log_level
from netqasm.runtime.application import default_app_instance
from netqasm.sdk.external import NetQASMConnection, Socket, simulate_application
from netqasm.sdk.qubit import Qubit


def run_alice():
    socket = Socket("Alice", "Bob")

    with NetQASMConnection("Alice") as alice:
        # define variable that is known only at runtime
        # (namely when we receive it from Bob)
        angle = Template("angle_from_bob")

        q = Qubit(alice)
        q.rot_X(n=angle, d=4)
        m = q.measure()

        # compile subroutine but do not yet send to backend
        subroutine = alice.compile()
        assert subroutine is not None

        # contains an operand called `angle_from_bob` that does not have a concrete value
        print(subroutine)

        # receive the value from Bob
        angle = int(socket.recv())

        # instantiate subroutine with Bob's value
        subroutine.instantiate(alice.app_id, arguments={"angle_from_bob": angle})

        # send to backend
        alice.commit_subroutine(subroutine)

        # alice should have done an X-rotation of angle pi -> qubit must be |1>
        assert int(m) == 1


def run_bob():
    socket = Socket("Bob", "Alice")
    angle = 16  # in multiples of pi/16
    socket.send(str(angle))


def test_precompilation():
    set_log_level("INFO")
    app_instance = default_app_instance(
        [
            ("Alice", run_alice),
            ("Bob", run_bob),
        ]
    )
    simulate_application(
        app_instance,
        use_app_config=False,
        enable_logging=False,
    )


if __name__ == "__main__":
    test_precompilation()



"""Abstractions for classical runtime values.

This module contains the `BaseFuture` class and its subclasses.
"""


from __future__ import annotations

import abc
from typing import TYPE_CHECKING, List, Optional, Union

from netqasm.lang import operand
from netqasm.lang.ir import GenericInstr, ICmd, Symbols
from netqasm.lang.operand import Address, ArrayEntry
from netqasm.lang.parsing import parse_address, parse_register
from netqasm.typedefs import T_Cmd
from netqasm.util.log import HostLine

if TYPE_CHECKING:
    from netqasm.lang import ir
    from netqasm.sdk import connection as sdkconn
    from netqasm.sdk.builder import Builder, SdkForEachContext, SdkIfContext

# Generic type for classical values (that may only get a value at runtime).
T_CValue = Union[int, "Future", "RegFuture"]


# Error types.
class NoValueError(RuntimeError):
    pass


class NonConstantIndexError(RuntimeError):
    pass


def as_int_when_value(cls):
    """A decorator for the `BaseFuture` class which makes is behave like an `int`
    when the property `value` is not `None`.
    """

    def wrap_method(method_name):
        """Return a new method for the class given a method name"""
        int_method = getattr(int, method_name)

        def new_method(self, *args, **kwargs):
            """Check if the value is set, otherwise raise an error"""
            value = self.value
            if value is None:
                raise NoValueError(
                    f"The object '{repr(self)}' has no value yet, "
                    "consider flusing the current subroutine"
                )
            try:
                if isinstance(*args, Future):
                    raise RuntimeError(
                        "Cannot directly do a binary operation on two Futures. "
                        "Convert them to `int`s first."
                    )
            except TypeError:
                pass
            return int_method(value, *args, **kwargs)

        return new_method

    method_names = [
        "__abs__",
        "__add__",
        "__and__",
        "__bool__",
        "__ceil__",
        "__divmod__",
        "__eq__",
        "__float__",
        "__floor__",
        "__floordiv__",
        "__ge__",
        "__gt__",
        "__hash__",
        "__int__",
        "__invert__",
        "__le__",
        "__lshift__",
        "__lt__",
        "__mod__",
        "__mul__",
        "__ne__",
        "__neg__",
        "__or__",
        "__pos__",
        "__pow__",
        "__radd__",
        "__rand__",
        "__rdivmod__",
        "__rfloordiv__",
        "__rlshift__",
        "__rmod__",
        "__rmul__",
        "__ror__",
        "__round__",
        "__rpow__",
        "__rrshift__",
        "__rshift__",
        "__rsub__",
        "__rtruediv__",
        "__rxor__",
        "__sub__",
        "__truediv__",
        "__xor__",
        "bit_length",
        "conjugate",
        "denominator",
        "imag",
        "numerator",
        "real",
        "to_bytes",
    ]
    for method_name in method_names:
        setattr(cls, method_name, wrap_method(method_name))
    return cls


@as_int_when_value
class BaseFuture(int):
    """Base class for Future-like objects.

    A Future represents a classical value that becomes available at some point
    in the future. At the moment, a Future always represents an integer value.

    Futures have a `value` property that is either `None` (when the value is not yet
    available), or has a concrete integer value.
    Executing a subroutine on the quantum node controller makes the value property go
    from `None` to a concrete value, granted that the subroutine sets the value of
    whatever the Future represents.
    When the `value` property has a concrete value, the Future behaves like an `int`
    and can then also be used in Python expressions just like integers.

    Typically, a Future instance is obtained as the result of an SDK function, and not
    directly instantiated in application code.
    For example, calling `measure()` on a Qubit returns a Future representing the
    measurement outcome. Only after the subroutine containing the operations has been
    executed by the quantum node controller (by flushing the subroutine), the
    measurement outcome will have an actual value, and the Future can be resolved into
    an `int`.
    """

    @classmethod
    def __new__(cls, *args, **kwargs):
        return int.__new__(cls, 0)

    def __init__(self, connection: sdkconn.BaseNetQASMConnection):
        self._value: Optional[int] = None
        self._connection: sdkconn.BaseNetQASMConnection = connection
        self._subrt_result: Optional[str] = None

    def __repr__(self):
        return f"{self.__class__} with value={self.value}"

    @property
    def subrt_result(self) -> str:
        if self._subrt_result is None:
            raise RuntimeError("subrt result not defined for this Future")
        return self._subrt_result

    @subrt_result.setter
    def subrt_result(self, result_name: str) -> None:
        self._subrt_result = result_name

    @property
    def builder(self) -> Builder:
        return self._connection.builder

    @property
    def value(self) -> Optional[int]:
        """Get the value of the future.
        If it's not set yet, `None` is returned."""
        if self._value is not None:
            return self._value
        else:
            return self._try_get_value()

    @abc.abstractmethod
    def _try_get_value(self) -> Optional[int]:
        raise NotImplementedError

    def add(
        self,
        other: Union[int, str, operand.Register, BaseFuture],
        mod: Optional[int] = None,
    ) -> None:
        """Add another value to this Future's value.

        The result is stored in this Future.

        Let the quantum node controller add a value to the value represented by this
        Future. The addition operation is compiled into a subroutine and is fully
        executed by the quantum node controller.
        This avoids the need to wait for a subroutine result (which resolves the
        Future's `value`) and then doing the addition on the Host.

        :param other: value to add to this Future's value
        :param mod: do the addition modulo `mod`
        """
        raise NotImplementedError(
            f"add is not implemented for {self.__class__.__name__}"
        )

    def if_eq(self, other: Optional[T_CValue]) -> SdkIfContext:
        return self.builder.sdk_new_if_context(
            condition=GenericInstr.BEQ, op0=self, op1=other
        )

    def if_ne(self, other: Optional[T_CValue]) -> SdkIfContext:
        return self.builder.sdk_new_if_context(
            condition=GenericInstr.BNE, op0=self, op1=other
        )

    def if_lt(self, other: Optional[T_CValue]) -> SdkIfContext:
        return self.builder.sdk_new_if_context(
            condition=GenericInstr.BLT, op0=self, op1=other
        )

    def if_ge(self, other: Optional[T_CValue]) -> SdkIfContext:
        return self.builder.sdk_new_if_context(
            condition=GenericInstr.BGE, op0=self, op1=other
        )

    def if_ez(self) -> SdkIfContext:
        return self.builder.sdk_new_if_context(
            condition=GenericInstr.BEZ, op0=self, op1=None
        )

    def if_nz(self) -> SdkIfContext:
        return self.builder.sdk_new_if_context(
            condition=GenericInstr.BNZ, op0=self, op1=None
        )


class Future(BaseFuture):
    """Represents a single array entry value that will become available in the future.

    See `BaseFuture` for more explanation about Futures.
    """

    @classmethod
    def __new__(cls, *args, **kwargs):
        return int.__new__(cls, 0)

    def __init__(
        self,
        connection: sdkconn.BaseNetQASMConnection,
        address: int,
        index: Union[int, Future, operand.Register, RegFuture],
    ):
        """Future constructor. Typically not used directly.

        :param connection: connection through which subroutines are sent that contain
            the array entry corresponding to this Future
        :param address: address of the array
        :param index: index in the array
        """
        super().__init__(connection=connection)
        self._address: int = address
        self._index: Union[int, Future, operand.Register, RegFuture] = index

    def __str__(self) -> str:
        value = self.value
        if value is None:
            return f"Future__@{self._address}[{self._index}]"
            # return (
            #     f"{self.__class__.__name__} to be stored in array with address "
            #     f"{self._address} at index {self._index}.\n"
            #     "To access the value, the subroutine must first be executed which can be done by flushing."
            # )
        else:
            return str(value)

    def _try_get_value(self) -> Optional[int]:
        if not isinstance(self._index, int):
            raise NonConstantIndexError("index is not constant and cannot be resolved")
        if self._connection.shared_memory is None:
            return None
        value = self._connection.shared_memory.get_array_part(
            address=self._address, index=self._index
        )
        if not isinstance(value, int) and value is not None:
            raise RuntimeError(
                f"Something went wrong: future value {value} is not an int or None"
            )
        if value is not None:
            self._value = value
        return value

    def add(
        self,
        other: Union[int, str, operand.Register, BaseFuture],
        mod: Optional[int] = None,
    ) -> None:
        if isinstance(other, str):
            other = parse_register(other)

        # Store self in a temporary register
        tmp_register = self.builder._mem_mgr.get_inactive_register(activate=True)
        load_commands = self.get_load_commands(tmp_register)
        store_commands = self._get_store_commands(tmp_register)

        other_tmp_register: Optional[operand.Register] = None
        other_operand: Union[int, operand.Register]  # empty declaration for type hints

        # If other is a Future, also load this into a temporary register
        if isinstance(other, Future):
            other_operand = self.builder._mem_mgr.get_inactive_register(activate=True)
            other_tmp_register = other_operand
            load_commands += other.get_load_commands(other_tmp_register)
            store_commands += other._get_store_commands(other_tmp_register)
        elif isinstance(other, operand.Register) or isinstance(other, int):
            other_operand = other
        else:
            raise NotImplementedError("for type {type(other)}")

        add_operands: List[ir.T_ProtoOperand] = [
            tmp_register,
            tmp_register,
            other_operand,
        ]
        if mod is None:
            add_instr = GenericInstr.ADD
        else:
            if not isinstance(mod, int):
                raise NotImplementedError
            add_instr = GenericInstr.ADDM
            add_operands.append(mod)

        commands = (
            load_commands
            + [
                ICmd(
                    instruction=add_instr,
                    operands=add_operands,
                )
            ]
            + store_commands
        )

        self.builder._mem_mgr.remove_active_register(tmp_register)
        if other_tmp_register is not None:
            self.builder._mem_mgr.remove_active_register(other_tmp_register)

        self.builder.subrt_add_pending_commands(commands)

    def get_load_commands(self, register: operand.Register) -> List[T_Cmd]:
        """Return a list of ProtoSubroutine commands for loading this Future into
        the specified register."""
        return self._get_access_commands(GenericInstr.LOAD, register)

    def _get_store_commands(self, register: operand.Register) -> List[T_Cmd]:
        return self._get_access_commands(GenericInstr.STORE, register)

    def _get_access_commands(
        self, instruction: GenericInstr, register: operand.Register
    ) -> List[T_Cmd]:
        assert (
            instruction == GenericInstr.LOAD or instruction == GenericInstr.STORE
        ), "Not an access instruction"
        commands = []
        if isinstance(self._index, Future):
            if self._connection is not self._index._connection:
                raise RuntimeError(
                    "Future-index must be from the same connection as the future itself"
                )
            tmp_register = self.builder._mem_mgr.get_inactive_register()
            # NOTE this might be many commands if the index is a future with a future index etc
            with self.builder._activate_register(tmp_register):
                access_index_cmds = self._index._get_access_commands(
                    instruction=GenericInstr.LOAD,
                    register=tmp_register,
                )
            commands += access_index_cmds
            index: Union[int, operand.Register] = tmp_register
        elif isinstance(self._index, RegFuture):
            assert (
                self._index.reg is not None
            ), "Trying to use RegFuture that has no value yet"
            index = self._index.reg
        elif isinstance(self._index, int) or isinstance(self._index, operand.Register):
            index = self._index
        else:
            raise TypeError(
                f"Cannot use type {type(self._index)} as index to load future"
            )
        address_entry = parse_address(
            f"{Symbols.ADDRESS_START}{self._address}[{index}]"
        )
        access_cmd = ICmd(
            instruction=instruction,
            operands=[
                register,
                address_entry,
            ],
        )
        commands.append(access_cmd)
        return commands

    def get_address_entry(self) -> ArrayEntry:
        """Convert this Future to an ArrayEntry object to be used an instruction
        operand."""
        if isinstance(self._index, RegFuture):
            assert self._index.reg is not None, (
                f"cannot use RegFuture {self._index} as array index since "
                f"it does not yet have a value"
            )
            return ArrayEntry(Address(self._address), self._index.reg)
        elif isinstance(self._index, int) or isinstance(self._index, operand.Register):
            return ArrayEntry(Address(self._address), self._index)
        else:
            assert False, f"index type {self._index} not supported"


class RegFuture(BaseFuture):
    """Represents a single register value that will become available in the future.

    See `BaseFuture` for more explanation about Futures.
    """

    def __init__(
        self,
        connection: sdkconn.BaseNetQASMConnection,
        reg: Optional[operand.Register] = None,
    ):
        """RegFuture constructor. Typically not used directly.

        :param connection: connection through which subroutines are sent that contain
            the array entry corresponding to this Future
        :param reg: specific NetQASM register that will hold the Future's value.
            If None, a suitable register is automatically used.
        """
        super().__init__(connection=connection)
        self._reg: Optional[operand.Register] = reg

    @property
    def reg(self) -> Optional[operand.Register]:
        return self._reg

    @reg.setter
    def reg(self, new_val: operand.Register) -> None:
        self._reg = new_val

    def __str__(self) -> str:
        assert self.reg is not None
        value = self.value
        if value is None:
            return f"RegFuture__{self.reg}"
            # return (
            #     f"{self.__class__.__name__} to be stored in reg {self.reg} "
            #     "To access the value, the subroutine must first be executed which can be done by flushing."
            # )
        else:
            return str(value)

    def _try_get_value(self) -> Optional[int]:
        assert self.reg is not None
        if self._connection.shared_memory is None:
            return None
        try:
            value = self._connection.shared_memory.get_register(self.reg)
        except KeyError:
            return None
        if value is not None:
            self._value = value
        return value

    def add(
        self,
        other: Union[int, str, operand.Register, BaseFuture],
        mod: Optional[int] = None,
    ) -> None:
        assert self.reg is not None
        if isinstance(other, str):
            other = parse_register(other)

        # Store self in a temporary register
        load_commands = []
        store_commands = []

        other_tmp_register: Optional[operand.Register] = None
        other_operand: Union[int, operand.Register]  # empty declaration for type hints

        # If other is a Future, also load this into a temporary register
        if isinstance(other, Future):
            other_operand = self.builder._mem_mgr.get_inactive_register(activate=True)
            other_tmp_register = other_operand
            load_commands += other.get_load_commands(other_tmp_register)
            store_commands += other._get_store_commands(other_tmp_register)
        elif isinstance(other, operand.Register) or isinstance(other, int):
            other_operand = other
        else:
            raise NotImplementedError("for type {type(other)}")

        add_operands: List[ir.T_ProtoOperand] = [
            self.reg,
            self.reg,
            other_operand,
        ]
        if mod is None:
            add_instr = GenericInstr.ADD
        else:
            if not isinstance(mod, int):
                raise NotImplementedError
            add_instr = GenericInstr.ADDM
            add_operands.append(mod)

        commands = (
            load_commands
            + [
                ICmd(
                    instruction=add_instr,
                    operands=add_operands,
                )
            ]
            + store_commands
        )

        if other_tmp_register is not None:
            self.builder._mem_mgr.remove_active_register(other_tmp_register)

        self.builder.subrt_add_pending_commands(commands)


class Array:
    """Wrapper around an array in Shared Memory.

    An `Array` instance provides methods to inspect and operate on an array that
    exists in shared memory.
    They are typically obtained as return values to certain SDK methods.

    Elements or slices of the array can be captured as Futures.
    """

    def __init__(
        self,
        connection: sdkconn.BaseNetQASMConnection,
        length: int,
        address: int,
        init_values: Optional[List[Optional[int]]] = None,
        lineno: Optional[HostLine] = None,
    ):
        """Array constructor. Typically not used directly.

        :param connection: connection of the application this array is part of
        :param length: length of the array
        :param address: address of the array
        :param init_values: initial values of the array. Must have length `length`.
        :param lineno: line number where the array is created in the Python source code
        """
        if init_values is not None:
            if not all((isinstance(x, int) or x is None) for x in init_values):
                raise TypeError("Array needs to consist of int's or None's")
            length = len(init_values)
        assert isinstance(length, int) and length > 0, f"{length} is not a valid length"
        self._connection: sdkconn.BaseNetQASMConnection = connection
        self._length: int = length
        self._address: int = address
        self._init_values: Optional[List[Optional[int]]] = init_values
        self._lineno: Optional[HostLine] = lineno

    @property
    def lineno(self) -> Optional[HostLine]:
        """What line in host application file initiated this array"""
        return self._lineno

    def __len__(self) -> int:
        return self._length

    def __getitem__(
        self, index: Union[int, slice]
    ) -> Union[None, int, List[Optional[int]]]:
        return self._connection.shared_memory.get_array_part(
            address=self._address, index=index
        )

    @property
    def address(self) -> int:
        return self._address

    @property
    def builder(self) -> Builder:
        return self._connection.builder

    def get_future_index(
        self, index: Union[int, str, operand.Register, RegFuture]
    ) -> Future:
        """Get a Future representing a particular array element"""
        if isinstance(index, str):
            index = parse_register(index)
        return Future(
            connection=self._connection,
            address=self._address,
            index=index,
        )

    def get_future_slice(self, s: slice) -> List[Future]:
        """Get a list of Futures each representing one element in a particular
        array slice"""
        range_args = []
        for attr in ["start", "stop", "step"]:
            x = getattr(s, attr)
            if x is not None:
                if not isinstance(x, int):
                    raise NotImplementedError(
                        "Future slices can only be specified by integers at this point, "
                        f"not {type(x)}"
                    )
                range_args.append(x)
        return [self.get_future_index(index) for index in range(*range_args)]

    def foreach(self) -> SdkForEachContext:
        """Create a context of code that gets called for each element in the array.

        Returns a future of the array value at the current index.

        Code inside the context *must* be compilable to NetQASM, that is,
        it should only contain quantum operations and/or classical values that are
        are stored in shared memory (arrays and registers).
        No classical communication is allowed.

        Example:

        .. code-block::

            with NetQASMConnection(app_name="alice") as alice:
                outcomes = alice.new_array(10)
                values = alice.new_array(
                    10,
                    init_values=[random.randint(0, 1) for _ in range(10)]
                )
                with values.foreach() as v:
                    q = Qubit(alice)
                    with v.if_eq(1):
                        q.H()
                    q.measure(future=outcomes.get_future_index(i))
        """
        return self.builder.sdk_new_foreach_context(array=self, return_index=False)

    def enumerate(self) -> SdkForEachContext:
        """Create a context of code that gets called for each element in the array
        and includes a counter.

        Returns a tuple (`index`, `future`) where `future` is a Future of the array
        value at the current index.

        Code inside the context *must* be compilable to NetQASM, that is,
        it should only contain quantum operations and/or classical values that are
        are stored in shared memory (arrays and registers).
        No classical communication is allowed.

        Example:

        .. code-block::

            with NetQASMConnection(app_name="alice") as alice:
                outcomes = alice.new_array(10)
                values = alice.new_array(
                    10,
                    init_values=[random.randint(0, 1) for _ in range(10)]
                )
                with values.enumerate() as (i, v):
                    q = Qubit(alice)
                    with v.if_eq(1):
                        q.H()
                    q.measure(future=outcomes.get_future_index(i))
        """
        return self.builder.sdk_new_foreach_context(array=self, return_index=True)

    def undefine(self) -> None:
        """Undefine (i.e. set to 'None') all elements in the array."""
        self.builder._build_cmds_undefine_array(self)



from typing import List

ALPHA_LOWER = "abcdefghijklmnopqrstuvwxyz"
ALPHA_CAPITAL = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
ALPHA_ALL = ALPHA_LOWER + ALPHA_CAPITAL
NUM = "0123456789"
ALPHA_NUM = ALPHA_ALL + NUM


def group_by_word(line: str, seperator=" ", brackets=None) -> List[str]:
    """Groups a string by words and contents within brackets"""
    line = line.strip()
    line += seperator  # This makes it easer to find the last word
    _assert_valid_seperator(seperator)
    if brackets is not None:
        _assert_valid_brackets(seperator, brackets)
        start_bracket, end_bracket = brackets
    words = []
    while len(line) > 0:
        if brackets is not None:
            first_seperator = line.find(seperator)
            first_start_bracket = line.find(start_bracket)
            if first_start_bracket != -1 and first_start_bracket < first_seperator:
                end_string = f"{end_bracket}{seperator}"
            else:
                end_string = seperator
        else:
            end_string = seperator
        # Find the closing string
        end = line.find(end_string)
        if end == -1:
            raise ValueError("Not a valid string, could not find a closing bracket")
        word = line[: end + len(end_string) - 1]
        words.append(word)
        line = line[end + len(end_string) :]
    return words


def is_variable_name(variable):
    if not isinstance(variable, str):
        return False  # Should be a string
    if variable[0] not in ALPHA_ALL:
        return False  # Should start with a letter
    if not set(variable) < set(ALPHA_NUM + "_"):
        return False  # Should only contain letters, digits and underscore
    return True


def is_number(number):
    # Allow for negative numbers
    if number.startswith("-"):
        number = number[1:]
    return len(number) > 0 and set(number) <= set(NUM)


def is_float(value):
    if not value.count(".") == 1:
        return False
    parts = value.split(".")
    # Both int_part and dec_part cannot be zero
    if all(len(part) == 0 for part in parts):
        return False
    return all(len(part) == 0 or is_number(part) for part in parts)


def rspaces(val, min_chars=4):
    val = str(val)
    return " " * (min_chars - len(val)) + val


def _assert_valid_seperator(seperator):
    if not isinstance(seperator, str):
        raise TypeError("seperator should be a string")
    if len(seperator) == 0:
        raise ValueError("seperator should contain at least one character")


def _assert_valid_brackets(seperator, brackets):
    if not isinstance(brackets, str):
        raise TypeError("brackets should be a string")
    if not (len(brackets) == 2 and len(set(brackets)) == 2):
        raise ValueError("brackets should be two unique characters")
    if seperator in brackets:
        raise ValueError("seperator should not be part of brackets")



import random

from netqasm.logging.glob import get_netqasm_logger
from netqasm.runtime.application import default_app_instance
from netqasm.sdk import Qubit
from netqasm.sdk.external import NetQASMConnection, simulate_application

logger = get_netqasm_logger()


def run_alice():
    num = 10
    init_values = [random.randint(0, 1) for _ in range(num)]
    loop_register = "R0"

    with NetQASMConnection("Alice") as alice:
        array = alice.new_array(init_values=init_values)
        outcomes = alice.new_array(length=num)

        def body(alice, _):
            q = Qubit(alice)
            with array.get_future_index(loop_register).if_eq(1):
                q.X()
            q.measure(future=outcomes.get_future_index(loop_register))

        alice.loop_body(body, stop=num, loop_register=loop_register)
    outcomes = list(outcomes)
    logger.debug(f"outcomes: {outcomes}")
    logger.debug(f"init_values: {init_values}")
    assert outcomes == init_values


def test_new_array():
    app_instance = default_app_instance(
        [
            ("Alice", run_alice),
        ]
    )
    simulate_application(app_instance, use_app_config=False, enable_logging=False)



import math
from typing import Any, Dict, Union

from netqasm.lang.operand import Template
from netqasm.logging.glob import set_log_level
from netqasm.runtime.application import default_app_instance
from netqasm.sdk.epr_socket import EPRSocket
from netqasm.sdk.external import NetQASMConnection, Socket, simulate_application
from netqasm.sdk.futures import Future, RegFuture

PRECOMPILE: bool = True


def run_client(alpha: float, theta1: float, r1: int, use_rsp: bool) -> Dict[str, Any]:
    epr_socket = EPRSocket("server")
    csocket = Socket("client", "server")
    conn = NetQASMConnection("client", epr_sockets=[epr_socket])

    with conn:
        outcome: Union[Future, RegFuture]

        if use_rsp:
            # Compute the final rotation angle based on theta1 to be a multiple of pi / 16 in the
            # range [0, 32)
            t1 = (16 + round(theta1 * 16 / math.pi)) % 32
            result = epr_socket.create_rsp(rotations_local=(0, 8, t1))[0]
            outcome = result.raw_measurement_outcome
        else:
            epr = epr_socket.create_keep()[0]
            epr.rot_Z(angle=theta1)
            epr.H()
            outcome = epr.measure(store_array=False)

    p1 = int(outcome)

    delta1 = alpha - theta1 + (p1 + r1) * math.pi

    csocket.send(str(delta1))

    return {"p1": p1}


def run_server(use_rsp: bool) -> Dict[str, Any]:
    epr_socket = EPRSocket("client")
    csocket = Socket("server", "client")
    conn = NetQASMConnection("server", epr_sockets=[epr_socket])

    with conn:
        if use_rsp:
            epr = epr_socket.recv_rsp()[0]
        else:
            epr = epr_socket.recv_keep()[0]

        conn.flush()

        if PRECOMPILE:
            epr.rot_Z(n=Template("delta1_numerator"), d=4)
            epr.H()
            outcome = epr.measure(store_array=False)
            subroutine = conn.compile()

            delta1 = float(csocket.recv())
            numerator = int(delta1 / (math.pi / 16))
            subroutine.instantiate(
                conn.app_id, arguments={"delta1_numerator": numerator}
            )
            conn.commit_subroutine(subroutine)

        else:
            delta1 = float(csocket.recv())
            epr.rot_Z(angle=delta1)

            epr.H()
            outcome = epr.measure(store_array=False)

    m2 = int(outcome)

    return {"m2": m2}


if __name__ == "__main__":
    set_log_level("INFO")

    app_instance = default_app_instance(
        [
            ("client", run_client),
            ("server", run_server),
        ]
    )

    use_rsp = False
    app_instance.program_inputs = {
        "client": {"alpha": 0, "theta1": 0, "r1": 0, "use_rsp": use_rsp},
        "server": {"use_rsp": use_rsp},
    }
    results = simulate_application(
        app_instance,
        use_app_config=False,
    )
    print(results)



from netqasm.logging.glob import set_log_level
from netqasm.sdk.connection import DebugConnection
from netqasm.sdk.qubit import Qubit


def main(no_output=False):
    with DebugConnection("Alice") as alice:
        num = 10

        outcomes = alice.new_array(num)
        even = alice.new_array(init_values=[0]).get_future_index(0)

        with alice.loop(num) as i:
            q = Qubit(alice)
            with even.if_eq(0):
                q.X()
            outcome = outcomes.get_future_index(i)
            q.measure(outcome)
            even.add(1, mod=2)

    if not no_output:
        print(f"binary:\n{alice.storage[0]}")


if __name__ == "__main__":
    set_log_level("INFO")
    main()



import sys

if ((sys.version_info[0] == 3) and (sys.version_info[1] >= 8)) or (
    sys.version_info[0] >= 4
):
    from typing import TypedDict
elif (sys.version_info[0] == 3) and (sys.version_info[1] in [6, 7]):
    from typing_extensions import TypedDict
else:
    raise RuntimeError("Unsupported Python version, {0}".format(sys.version_info))



# qlink-layer 0.1.0

from collections import namedtuple
from enum import Enum, auto
from typing import Union

import qlink_interface as qlink_1_0


class EPRType(Enum):
    K = 0
    M = auto()
    R = auto()


class EPRRole(Enum):
    CREATE = 0
    RECV = auto()


# Supported request types (create and keep, measure directly, and remote state preparation)
class RequestType(Enum):
    K = EPRType.K.value
    M = EPRType.M.value
    R = EPRType.R.value
    RECV = auto()
    STOP_RECV = auto()


# Types of replies from the link layer protocol
class ReturnType(Enum):
    OK_K = EPRType.K.value
    OK_M = EPRType.M.value
    OK_R = EPRType.R.value
    ERR = auto()
    CREATE_ID = auto()


# Error messages
class ErrorCode(Enum):
    UNSUPP = 0
    NOTIME = auto()
    NORES = auto()
    TIMEOUT = auto()
    REJECTED = auto()
    OTHER = auto()
    EXPIRE = auto()
    CREATE = auto()


# Choice of random bases sets
class RandomBasis(Enum):
    NONE = 0
    XZ = auto()
    XYZ = auto()
    CHSH = auto()


# Bases return from link layer about which basis was randomly chosen
class Basis(Enum):
    Z = 0
    X = auto()
    Y = auto()
    ZPLUSX = auto()
    ZMINUSX = auto()


# What Bell state is generated
class BellState(Enum):
    PHI_PLUS = 0  # |00> + |11>
    PSI_PLUS = 1  # |01> + |10>
    PSI_MINUS = 2  # |01> - |10>
    PHI_MINUS = 3  # |00> - |11>


# Unit for max_time argument
class TimeUnit(Enum):
    MICRO_SECONDS = 0
    MILLI_SECONDS = 1
    SECONDS = 2


# CREATE message to the link layer for entanglement generation
LinkLayerCreate = namedtuple(
    "LinkLayerCreate",
    [
        "remote_node_id",
        "purpose_id",
        "type",
        "number",
        "random_basis_local",
        "random_basis_remote",
        "minimum_fidelity",
        "time_unit",
        "max_time",
        "priority",
        "atomic",
        "consecutive",
        "probability_dist_local1",
        "probability_dist_local2",
        "probability_dist_remote1",
        "probability_dist_remote2",
        "rotation_X_local1",
        "rotation_Y_local",
        "rotation_X_local2",
        "rotation_X_remote1",
        "rotation_Y_remote",
        "rotation_X_remote2",
    ],
)
LinkLayerCreate.__new__.__defaults__ = (  # type: ignore
    0,
    0,
    RequestType.K,
    1,
    RandomBasis.NONE,
    RandomBasis.NONE,
) + (0,) * (len(LinkLayerCreate._fields) - 6)

# RECV message to the link layer to allow for entanglement generation with a remote node
LinkLayerRecv = namedtuple(
    "LinkLayerRecv",
    [
        "type",
        "remote_node_id",
        "purpose_id",
    ],
)
LinkLayerRecv.__new__.__defaults__ = (RequestType.RECV,) + (0,) * (  # type: ignore
    len(LinkLayerRecv._fields) - 1
)

# RECV message to the link layer to stop allowing for entanglement generation with a remote node
LinkLayerStopRecv = namedtuple(
    "LinkLayerStopRecv",
    [
        "type",
        "remote_node_id",
        "purpose_id",
    ],
)
LinkLayerStopRecv.__new__.__defaults__ = (RequestType.STOP_RECV,) + (0,) * (  # type: ignore
    len(LinkLayerStopRecv._fields) - 1
)

# OK message from the link layer of successful generation of entanglement that is kept in memory
LinkLayerOKTypeK = namedtuple(
    "LinkLayerOKTypeK",
    [
        "type",
        "create_id",
        "logical_qubit_id",
        "directionality_flag",
        "sequence_number",
        "purpose_id",
        "remote_node_id",
        "goodness",
        "goodness_time",
        "bell_state",
    ],
)
LinkLayerOKTypeK.__new__.__defaults__ = (ReturnType.OK_K,) + (0,) * (  # type: ignore
    len(LinkLayerOKTypeK._fields) - 1
)

# OK message from the link layer of successful generation of entanglement that is measured directly
LinkLayerOKTypeM = namedtuple(
    "LinkLayerOKTypeM",
    [
        "type",
        "create_id",
        "measurement_outcome",
        "measurement_basis",
        "directionality_flag",
        "sequence_number",
        "purpose_id",
        "remote_node_id",
        "goodness",
        "bell_state",
    ],
)
LinkLayerOKTypeM.__new__.__defaults__ = (ReturnType.OK_M,) + (0,) * (  # type: ignore
    len(LinkLayerOKTypeM._fields) - 1
)

# OK message from the link layer of successful generation of entanglement that was used for remote state preparation
# (to creator node)
LinkLayerOKTypeR = namedtuple(
    "LinkLayerOKTypeR",
    [
        "type",
        "create_id",
        "measurement_outcome",
        "directionality_flag",
        "sequence_number",
        "purpose_id",
        "remote_node_id",
        "goodness",
        "bell_state",
    ],
)
LinkLayerOKTypeR.__new__.__defaults__ = (ReturnType.OK_R,) + (0,) * (  # type: ignore
    len(LinkLayerOKTypeR._fields) - 1
)

# Error message from the link layer
LinkLayerErr = namedtuple(
    "LinkLayerErr",
    [
        "type",
        "create_id",
        "error_code",
        "use_sequence_number_range",
        "sequence_number_low",
        "sequence_number_high",
        "origin_node_id",
    ],
)
LinkLayerErr.__new__.__defaults__ = (ReturnType.ERR,) + (0,) * (  # type: ignore
    len(LinkLayerErr._fields) - 1
)


def get_creator_node_id(local_node_id, create_request):
    """Returns the node ID of the node that submitted the given create request"""
    if create_request.directionality_flag == 1:
        return create_request.remote_node_id
    else:
        return local_node_id


T_LinkLayerResponse = Union[
    LinkLayerOKTypeK, LinkLayerOKTypeM, LinkLayerOKTypeR, LinkLayerErr
]

T_LinkLayer_1_0_Response = Union[
    qlink_1_0.ResCreateAndKeep, qlink_1_0.ResMeasureDirectly, qlink_1_0.ResError
]


def request_to_qlink_1_0(
    request: Union[LinkLayerCreate, LinkLayerRecv]
) -> qlink_1_0.ReqCreateBase:
    if request.type == RequestType.K:
        assert isinstance(request, LinkLayerCreate)
        return qlink_1_0.ReqCreateAndKeep(
            remote_node_id=request.remote_node_id,
            minimum_fidelity=request.minimum_fidelity,
            time_unit=request.time_unit,
            max_time=request.max_time,
            purpose_id=request.purpose_id,
            number=request.number,
            priority=request.priority,
            atomic=request.atomic,
            consecutive=request.consecutive,
        )

    elif request.type == RequestType.M:
        assert isinstance(request, LinkLayerCreate)
        return qlink_1_0.ReqMeasureDirectly(
            remote_node_id=request.remote_node_id,
            minimum_fidelity=request.minimum_fidelity,
            time_unit=request.time_unit,
            max_time=request.max_time,
            purpose_id=request.purpose_id,
            number=request.number,
            priority=request.priority,
            atomic=request.atomic,
            consecutive=request.consecutive,
            random_basis_local=qlink_1_0.RandomBasis(request.random_basis_local.value),
            random_basis_remote=qlink_1_0.RandomBasis(
                request.random_basis_remote.value
            ),
            x_rotation_angle_local_1=request.rotation_X_local1,
            y_rotation_angle_local=request.rotation_Y_local,
            x_rotation_angle_local_2=request.rotation_X_local2,
            x_rotation_angle_remote_1=request.rotation_X_remote1,
            y_rotation_angle_remote=request.rotation_Y_remote,
            x_rotation_angle_remote_2=request.rotation_X_remote2,
            probability_distribution_parameter_local_1=request.probability_dist_local1,
            probability_distribution_parameter_remote_1=request.probability_dist_remote1,
            probability_distribution_parameter_local_2=request.probability_dist_local2,
            probability_distribution_parameter_remote_2=request.probability_dist_remote2,
        )
    elif request.type == RequestType.RECV:
        assert isinstance(request, LinkLayerRecv)
        return qlink_1_0.ReqReceive(
            remote_node_id=request.remote_node_id, purpose_id=request.purpose_id
        )
    else:
        raise ValueError(f"Cannot convert request {request} to qlink-interface 1.0")


def response_from_qlink_1_0(response: T_LinkLayer_1_0_Response) -> T_LinkLayerResponse:
    if isinstance(response, qlink_1_0.ResCreateAndKeep):
        return LinkLayerOKTypeK(
            type=ReturnType.OK_K,
            create_id=response.create_id,
            logical_qubit_id=response.logical_qubit_id,
            directionality_flag=response.directionality_flag,
            sequence_number=response.sequence_number,
            purpose_id=response.purpose_id,
            remote_node_id=response.remote_node_id,
            goodness=response.goodness,
            goodness_time=response.time_of_goodness,
            bell_state=response.bell_state,
        )
    elif isinstance(response, qlink_1_0.ResMeasureDirectly):
        return LinkLayerOKTypeM(
            type=ReturnType.OK_M,
            create_id=response.create_id,
            measurement_outcome=response.measurement_outcome,
            measurement_basis=Basis(response.measurement_basis.value),
            directionality_flag=response.directionality_flag,
            sequence_number=response.sequence_number,
            purpose_id=response.purpose_id,
            remote_node_id=response.remote_node_id,
            goodness=response.goodness,
            bell_state=response.bell_state,
        )
    elif isinstance(response, qlink_1_0.ResError):
        return LinkLayerErr(
            type=ReturnType.ERR,
            create_id=response.create_id,
            error_code=response.error_code,
            use_sequence_number_range=response.use_sequence_number_range,
            sequence_number_low=response.sequence_number_low,
            sequence_number_high=response.sequence_number_high,
            origin_node_id=response.origin_node_id,
        )
    else:
        raise ValueError(f"Cannot convert response {response} from qlink-interface 1.0")



import math

from numpy import linalg

from netqasm.logging.output import get_new_app_logger
from netqasm.runtime.settings import Simulator, get_simulator
from netqasm.sdk import EPRSocket, Qubit
from netqasm.sdk.external import NetQASMConnection, Socket, get_qubit_state
from netqasm.sdk.toolbox import set_qubit_state
from netqasm.sdk.toolbox.sim_states import qubit_from, to_dm


def main(app_config=None, phi=0.0, theta=0.0):
    # Inputs are coefficients of pi, e.g. phi=0.5 -> angle 0.5*pi
    phi *= math.pi
    theta *= math.pi

    log_config = app_config.log_config

    # socket for creating an EPR pair with target
    target_epr = EPRSocket("target")

    # socket for communicating classical messages with target
    class_socket = Socket("controller", "target", log_config=app_config.log_config)

    app_logger = get_new_app_logger(app_name="sender", log_config=log_config)

    # connect to the back-end
    controller = NetQASMConnection(
        app_name=app_config.app_name,
        log_config=app_config.log_config,
        epr_sockets=[target_epr],
    )

    with controller:
        app_logger.log("Creating EPR pair with target...")
        # create one EPR pair with target
        epr = target_epr.create_keep(1)[0]

        # initialize control qubit of the distributed CNOT
        controller.flush()
        app_logger.log("Initializing control qubit...")
        ctrl_qubit = Qubit(controller)
        set_qubit_state(ctrl_qubit, phi, theta)
        controller.flush()
        app_logger.log("Initialized control qubit")

        # Synchronize with other nodes for cleaner logs/animations in QNE.
        class_socket.recv_silent()

        app_logger.log("Starting distributed CNOT...")
        # perform a local CNOT with `epr` and measure `epr`
        ctrl_qubit.cnot(epr)
        epr_meas = epr.measure()

        # let back-end execute the quantum operations above
        controller.flush()

        # send the outcome to target
        class_socket.send(str(epr_meas))

        # wait for target's measurement outcome to undo potential entanglement
        # between his EPR half and the original control qubit
        target_meas = class_socket.recv()
        if target_meas == "1":
            app_logger.log("Outcome = 1, so doing Z correction")
            ctrl_qubit.Z()
        else:
            app_logger.log("Outcome = 0, no corrections needed")

        controller.flush()

        # ack the outcome
        class_socket.send_silent("ACK")

        # Get information about the final state. Only possible in simulation and
        # for visualizations purposes in QNE.
        if get_simulator() == Simulator.NETSQUID:
            original_dm = to_dm(qubit_from(phi, theta))
            final_dm = get_qubit_state(ctrl_qubit, reduced_dm=True)

            if linalg.matrix_rank(final_dm) != 1:
                entangled_text = (
                    "The controller's and target's qubits are entangled "
                    "so their individual states cannot be shown."
                )
            else:
                entangled_text = (
                    "The controller's and target's qubits are *not* "
                    "entangled. Their final states are displayed below."
                )
        else:
            original_dm = None
            final_dm = None
            entangled_text = ""

    return {
        "epr_meas": int(epr_meas),
        "original_state": original_dm.tolist(),
        "final_state": final_dm if final_dm is None else final_dm.tolist(),
        "entangled_text": entangled_text,
    }



import os
import pickle
import shutil

from netqasm.sdk.connection import BaseNetQASMConnection
from netqasm.util.yaml import dump_yaml, load_yaml

_LAST_LOG = "LAST"


def process_log(log_dir):
    # Add host line numbers to logs
    _add_hln_to_logs(log_dir)
    _create_app_instr_logs(log_dir)
    make_last_log(log_dir)


def make_last_log(log_dir):
    # Make this the last log
    base_log_dir, _log_dir_name = os.path.split(log_dir)
    last_log_dir = os.path.join(base_log_dir, _LAST_LOG)
    if os.path.exists(last_log_dir):
        shutil.rmtree(last_log_dir)
    shutil.copytree(log_dir, last_log_dir)


def _add_hln_to_logs(log_dir):
    file_end = "_instrs.yaml"
    for entry in os.listdir(log_dir):
        if entry.endswith(file_end):
            node_name = entry[: -len(file_end)]
            output_file_path = os.path.join(log_dir, entry)
            subroutines_file_path = os.path.join(
                log_dir, f"subroutines_{node_name}.pkl"
            )
            _add_hln_to_log(
                output_file_path=output_file_path,
                subroutines_file_path=subroutines_file_path,
            )


def _add_hln_to_log(output_file_path, subroutines_file_path):
    if not os.path.exists(subroutines_file_path):
        return

    # Read subroutines and log file
    with open(subroutines_file_path, "rb") as f:
        subroutines = pickle.load(f)
    data = load_yaml(output_file_path)

    # Update entries
    for entry in data:
        _add_hln_to_log_entry(subroutines, entry)

    # Write updated log file
    dump_yaml(data, output_file_path)


def _add_hln_to_log_entry(subroutines, entry):
    prc = entry["PRC"]
    sid = entry["SID"]
    subroutine = subroutines[sid]
    hostline = subroutine.commands[prc].lineno
    entry["HLN"] = hostline.lineno
    entry["HFL"] = hostline.filename


def _create_app_instr_logs(log_dir):
    file_end = "_instrs.yaml"

    app_names = BaseNetQASMConnection.get_app_names()

    for entry in os.listdir(log_dir):
        if entry.endswith(file_end):
            node_name = entry[: -len(file_end)]

            if node_name not in app_names.keys():
                raise ValueError(
                    f"Node {node_name} has logged instructions, but no app ID is known to have run on this node."
                )

            if len(app_names[node_name]) > 1:
                raise ValueError(
                    "Logging does not currently support multiple apps per node."
                )

            app_name = list(app_names[node_name].values())[0]

            node_instr_log_file = os.path.join(log_dir, entry)
            app_instr_log_file = os.path.join(log_dir, f"{app_name}_instrs.yaml")

            # TODO
            # Create an {app_name}_instrs.yaml file for each app_id found in {node_name}_instrs.yaml
            # Currently, it expects all app IDs to be the same, so we can just rename the file

            os.rename(node_instr_log_file, app_instr_log_file)



from netqasm.logging.glob import get_netqasm_logger
from netqasm.qlink_compat import BellState
from netqasm.runtime.application import default_app_instance
from netqasm.sdk import EPRSocket
from netqasm.sdk.external import NetQASMConnection, simulate_application

logger = get_netqasm_logger()

num = 10


def run_alice():
    epr_socket = EPRSocket("bob")
    outcomes = []
    with NetQASMConnection("alice", epr_sockets=[epr_socket]):
        ent_infos = epr_socket.create_measure(number=num)
        for ent_info in ent_infos:
            outcomes.append(ent_info.raw_measurement_outcome)
    return outcomes


def run_bob():
    epr_socket = EPRSocket("alice")
    outcomes = []
    with NetQASMConnection("bob", epr_sockets=[epr_socket]):
        ent_infos = epr_socket.recv_measure(number=num)
        for ent_info in ent_infos:
            outcomes.append(ent_info.raw_measurement_outcome)
    return outcomes


def test_create_epr_m():
    app_instance = default_app_instance(
        [
            ("alice", run_alice),
            ("bob", run_bob),
        ]
    )
    outcomes = simulate_application(
        app_instance, use_app_config=False, enable_logging=False
    )[0]

    alice_outcomes = [int(m) for m in outcomes["app_alice"]]
    bob_outcomes = [int(m) for m in outcomes["app_bob"]]
    print(alice_outcomes)
    print(bob_outcomes)
    assert all(i == j for i, j in zip(alice_outcomes, bob_outcomes))


def run_alice_psi_plus():
    epr_socket = EPRSocket("bob")
    outcomes = []

    with NetQASMConnection("alice", epr_sockets=[epr_socket]):
        results = epr_socket.create_measure(number=num)

    for result in results:
        # Pretend we had Psi+ instead of Phi+.
        # Post-processing should make up for this.
        result.raw_bell_state._value = BellState.PSI_PLUS.value
        result.post_process = True
        outcomes.append(result.measurement_outcome)
    return outcomes


def run_bob_psi_plus():
    epr_socket = EPRSocket("alice")
    outcomes = []

    with NetQASMConnection("bob", epr_sockets=[epr_socket]):
        results = epr_socket.recv_measure(number=num)

    for result in results:
        # Pretend we had Psi+ instead of Phi+ by flipping the bit
        outcomes.append(int(result.raw_measurement_outcome) ^ 1)
    return outcomes


def test_create_epr_m_psi_plus():
    app_instance = default_app_instance(
        [
            ("alice", run_alice_psi_plus),
            ("bob", run_bob_psi_plus),
        ]
    )
    outcomes = simulate_application(
        app_instance, use_app_config=False, enable_logging=False
    )[0]

    alice_outcomes = [int(m) for m in outcomes["app_alice"]]
    bob_outcomes = [int(m) for m in outcomes["app_bob"]]
    print(alice_outcomes)
    print(bob_outcomes)
    assert all(i == j for i, j in zip(alice_outcomes, bob_outcomes))


if __name__ == "__main__":
    test_create_epr_m()
    test_create_epr_m_psi_plus()



import json
import math
import random
from dataclasses import dataclass
from typing import Optional

from netqasm.logging.glob import get_netqasm_logger
from netqasm.sdk import EPRSocket
from netqasm.sdk.classical_communication.message import StructuredMessage
from netqasm.sdk.external import NetQASMConnection, Socket

logger = get_netqasm_logger()

buf_msgs = []  # type: ignore
EOF = "EOF"
ALL_MEASURED = "All qubits measured"


def recv_single_msg(socket):
    """Used to not get multiple messages at a time"""
    if len(buf_msgs) > 0:
        msg = buf_msgs.pop(0)
    else:
        msgs = socket.recv().split(EOF)[:-1]
        buf_msgs.extend(msgs[1:])
        msg = msgs[0]
    logger.debug(f"Alice received msg {msg}")
    return msg


def send_single_msg(socket, msg):
    """Used to not get multiple messages at a time"""
    socket.send(msg + EOF)


def sendClassicalAssured(socket, data):
    data = json.dumps(data)
    send_single_msg(socket, data)
    while recv_single_msg(socket) != "ACK":
        pass


def recvClassicalAssured(socket):
    data = recv_single_msg(socket)
    data = json.loads(data)
    send_single_msg(socket, "ACK")
    return data


def distribute_bb84_states(conn, epr_socket, socket, target, n):
    bit_flips = [None for _ in range(n)]
    basis_flips = [random.randint(0, 1) for _ in range(n)]

    for i in range(n):
        q = epr_socket.create_keep(1)[0]
        if basis_flips[i]:
            q.H()
        m = q.measure()
        conn.flush()
        # Synchronize with the other node so that entanglement operations
        # appear more cleanly in the logs (not needed in principle).
        socket.send_silent("sync")
        socket.recv_silent()
        bit_flips[i] = int(m)
    return bit_flips, basis_flips


def filter_bases(socket, pairs_info):
    bases = [(i, pairs_info[i].basis) for (i, pair) in enumerate(pairs_info)]

    msg = StructuredMessage(header="Bases", payload=bases)
    socket.send_structured(msg)
    remote_bases = socket.recv_structured().payload

    for (i, basis), (remote_i, remote_basis) in zip(bases, remote_bases):
        assert i == remote_i
        pairs_info[i].same_basis = basis == remote_basis

    return pairs_info


def estimate_error_rate(socket, pairs_info, num_test_bits):
    same_basis_indices = [pair.index for pair in pairs_info if pair.same_basis]
    test_indices = random.sample(
        same_basis_indices, min(num_test_bits, len(same_basis_indices))
    )
    for pair in pairs_info:
        pair.test_outcome = pair.index in test_indices

    test_outcomes = [(i, pairs_info[i].outcome) for i in test_indices]

    logger.info(f"alice finding {num_test_bits} test bits")
    logger.info(f"alice test indices: {test_indices}")
    logger.info(f"alice test outcomes: {test_outcomes}")

    socket.send_structured(StructuredMessage("Test indices", test_indices))
    target_test_outcomes = socket.recv_structured().payload
    socket.send_structured(StructuredMessage("Test outcomes", test_outcomes))
    logger.info(f"alice target_test_outcomes: {target_test_outcomes}")

    num_error = 0
    for (i1, t1), (i2, t2) in zip(test_outcomes, target_test_outcomes):
        assert i1 == i2
        if t1 != t2:
            num_error += 1
            pairs_info[i1].same_outcome = False
        else:
            pairs_info[i1].same_outcome = True

    return pairs_info, (num_error / num_test_bits)


def extract_key(x, r):
    return sum([xj * rj for xj, rj in zip(x, r)]) % 2


def h(p):
    if p == 0 or p == 1:
        return 0
    else:
        return -p * math.log2(p) - (1 - p) * math.log2(1 - p)


@dataclass
class PairInfo:
    """Information that Alice has about one generated pair.
    The information is filled progressively during the protocol."""

    # Index in list of all generated pairs.
    index: int

    # Basis Alice measured in. 0 = Z, 1 = X.
    basis: int

    # Measurement outcome (0 or 1).
    outcome: int

    # Whether Bob measured his qubit in the same basis or not.
    same_basis: Optional[bool] = None

    # Whether to use this pair to estimate errors by comparing the outcomes.
    test_outcome: Optional[bool] = None

    # Whether measurement outcome is the same as Bob's. (Only for pairs used for error estimation.)
    same_outcome: Optional[bool] = None


def main(app_config=None, num_bits=100):
    num_test_bits = max(num_bits // 4, 1)

    # Socket for classical communication
    socket = Socket("alice", "bob", log_config=app_config.log_config)
    # Socket for EPR generation
    epr_socket = EPRSocket("bob")

    alice = NetQASMConnection(
        app_name=app_config.app_name,
        log_config=app_config.log_config,
        epr_sockets=[epr_socket],
    )
    with alice:
        bit_flips, basis_flips = distribute_bb84_states(
            alice, epr_socket, socket, "bob", num_bits
        )

    outcomes = [int(b) for b in bit_flips]
    theta = [int(b) for b in basis_flips]

    logger.info(f"alice outcomes: {outcomes}")
    logger.info(f"alice theta: {theta}")

    pairs_info = []
    for i in range(num_bits):
        pairs_info.append(
            PairInfo(
                index=i,
                basis=int(basis_flips[i]),
                outcome=int(bit_flips[i]),
            )
        )

    m = socket.recv()
    if m != ALL_MEASURED:
        logger.info(m)
        raise RuntimeError("Failed to distribute BB84 states")

    pairs_info = filter_bases(socket, pairs_info)

    pairs_info, error_rate = estimate_error_rate(socket, pairs_info, num_test_bits)
    logger.info(f"alice error rate: {error_rate}")

    raw_key = [
        pair.outcome for pair in pairs_info if pair.same_basis and not pair.test_outcome
    ]
    logger.info(f"alice raw key: {raw_key}")

    if len(raw_key) == 0:
        raw_key_text = "empty - there were not enough outcomes left"
    else:
        raw_key_text = str(raw_key)

    print(f"alice raw key: {raw_key}")

    # Return data.

    table = []
    for pair in pairs_info:
        basis = "X" if pair.basis == 1 else "Z"
        check = pair.same_outcome if pair.test_outcome else "-"
        table.append([pair.index, basis, pair.same_basis, pair.outcome, check])

    x_basis_count = sum(pair.basis for pair in pairs_info)
    z_basis_count = num_bits - x_basis_count
    same_basis_count = sum(pair.same_basis for pair in pairs_info)
    outcome_comparison_count = sum(
        pair.test_outcome for pair in pairs_info if pair.same_basis
    )
    diff_outcome_count = outcome_comparison_count - sum(
        pair.same_outcome for pair in pairs_info if pair.test_outcome
    )
    if outcome_comparison_count == 0:
        qber = 1
    else:
        qber = (diff_outcome_count) / outcome_comparison_count
    key_rate_potential = 1 - 2 * h(qber)

    return {
        # Table with one row per generated pair.
        # Columns:
        #   - Pair number
        #   - Measurement basis ("X" or "Z")
        #   - Same basis as Bob ("True" or "False")
        #   - Measurement outcome ("0" or "1")
        #   - Outcome same as Bob ("True", "False" or "-")
        #       ("-" is when outcomes are not compared)
        "table": table,
        # Number of times measured in the X basis.
        "x_basis_count": x_basis_count,
        # Number of times measured in the Z basis.
        "z_basis_count": z_basis_count,
        # Number of times measured in the same basis as Bob.
        "same_basis_count": same_basis_count,
        # Number of pairs chosen to compare measurement outcomes for.
        "outcome_comparison_count": outcome_comparison_count,
        # Number of compared outcomes with equal values.
        "diff_outcome_count": diff_outcome_count,
        # Estimated Quantum Bit Error Rate (QBER).
        "qber": qber,
        # Rate of secure key that can in theory be extracted from the raw key.
        "key_rate_potential": key_rate_potential,
        # Raw key.
        # ('Result' of this application. In practice, there'll be post-processing to produce secure shared key.)
        "raw_key": raw_key_text,
    }


if __name__ == "__main__":
    main()



from dataclasses import dataclass
from typing import Any, Callable, Dict, Optional

from netqasm.sdk.config import LogConfig


@dataclass
class AppConfig:
    app_name: str
    node_name: str
    main_func: Callable
    log_config: Optional[LogConfig]
    inputs: Dict[str, Any]


def default_app_config(app_name: str, main_func: Callable) -> AppConfig:
    return AppConfig(
        app_name=app_name,
        node_name=app_name,
        main_func=main_func,
        log_config=None,
        inputs={},
    )



import numpy as np
from scipy import linalg

from netqasm.lang.ir import GenericInstr

# Single-qubit gates
f = np.sqrt(1 / 2)
X = np.array([[0, 1], [1, 0]])
Y = np.array([[0, -1j], [1j, 0]])
Z = np.array([[1, 0], [0, -1]])
PAULIS = [X, Y, Z]
H = (X + Z) * f
K = (Y + Z) * f
S = np.array([[1, 0], [0, 1j]])
T = np.array([[1, 0], [0, (1 + 1j) * f]])
# Two-qubit gates
CNOT = np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]])
CPHASE = np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, -1]])


STATIC_QUBIT_GATE_TO_MATRIX = {
    GenericInstr.X: X,
    GenericInstr.Y: Y,
    GenericInstr.Z: Z,
    GenericInstr.H: H,
    GenericInstr.K: K,
    GenericInstr.S: S,
    GenericInstr.T: T,
    GenericInstr.CNOT: CNOT,
    GenericInstr.CPHASE: CPHASE,
}


def get_rotation_matrix(axis, angle) -> np.ndarray:
    """Returns a single-qubit rotation matrix given an axis and an angle"""
    norm = linalg.norm(axis)
    if norm == 0:
        raise ValueError("Axis need to have non-negative norm")
    axis = axis / norm
    rot_mat = linalg.expm(-1j * angle / 2 * sum(a * P for a, P in zip(axis, PAULIS)))
    return rot_mat  # type: ignore


def get_controlled_rotation_matrix(axis, angle) -> np.ndarray:
    target_pos = get_rotation_matrix(axis, angle)
    target_neg = get_rotation_matrix(axis, -angle)

    ctrl_zero = np.array([[1, 0], [0, 0]])
    ctrl_one = np.array([[0, 0], [0, 1]])

    inv_controlled_gate = np.kron(ctrl_zero, target_pos) + np.kron(ctrl_one, np.eye(2))
    controlled_gate = np.kron(ctrl_one, target_neg) + np.kron(ctrl_zero, np.eye(2))

    return inv_controlled_gate @ controlled_gate


def gate_to_matrix(instr, angle=None):
    """Returns the matrix representation of a quantum gate"""
    if instr in STATIC_QUBIT_GATE_TO_MATRIX:
        return STATIC_QUBIT_GATE_TO_MATRIX[instr]
    elif instr in [GenericInstr.ROT_X, GenericInstr.ROT_Y, GenericInstr.ROT_Z]:
        if angle is None:
            raise TypeError(
                "To get the matrix of a rotation an angle needs to be specified"
            )
        axis = {
            GenericInstr.ROT_X: [1, 0, 0],
            GenericInstr.ROT_Y: [0, 1, 0],
            GenericInstr.ROT_Z: [0, 0, 1],
        }[instr]
        if isinstance(angle, tuple):
            n, d = angle
            angle = n * np.pi / 2**d
        return get_rotation_matrix(axis=axis, angle=angle)
    else:
        raise ValueError(f"{instr} is not a quantum gate")


def are_matrices_equal(*matrices):
    """Checks if two matrices are equal, disregarding any global phase"""
    if len(matrices) <= 1:
        return True
    # Get the first non zero entry of the first matrix
    non_zero_indices = np.nonzero(matrices[0])
    # Check if all entries are zero
    if len(non_zero_indices[0]) == 0:
        # All other matrices should also be zero
        for matrix in matrices[1:]:
            if not np.allclose(matrix, 0):
                return False
        return True
    first_non_zero = non_zero_indices[0][0], non_zero_indices[1][0]
    # Check equality against the rest of the matrices
    for matrix in matrices[1:]:
        # If another matrix has zero at this place, it is not equal
        if matrix[first_non_zero] == 0:
            return False
        # Check what phase to apply to the matrix so that it is the same as the first
        phase = np.angle(matrices[0][first_non_zero] / matrix[first_non_zero])
        if not np.allclose(matrices[0], np.exp(phase * 1j) * matrix):
            return False
    return True



"""Implementations of classical messaging interfaces using inter-thread communication.

These implementations are most suitable for simulators that run Hosts (nodes) in
separate threads, and where communication between Hosts is done by inter-thread
messaging.
"""

from .broadcast_channel import ThreadBroadcastChannel
from .socket import ThreadSocket
from .socket_hub import reset_socket_hub



from __future__ import annotations

from typing import TYPE_CHECKING, List, Tuple

import numpy as np

from netqasm.lang.encoding import IMMEDIATE_BITS

if TYPE_CHECKING:
    from netqasm.sdk import qubit


def set_qubit_state(qubit: qubit.Qubit, phi: float = 0.0, theta: float = 0.0) -> None:
    r"""Assuming that the qubit is in the state :math:`|0\rangle`, this function
    rotates the state to :math:`\cos(\theta / 2)|0\rangle + e^{i\phi}\sin(\theta / 2)|1\rangle`.

    Parameters
    ----------
    qubit : :class:`.sdk.qubit.Qubit`
        The qubit to prepare the state.
    phi : float
        Angle around Z-axis from X-axis
    theta : float
        Angle from Z-axis
    """
    qubit.rot_Y(angle=theta)
    qubit.rot_Z(angle=phi)


def get_angle_spec_from_float(angle: float, tol: float = 1e-4) -> List[Tuple[int, int]]:
    r"""Tries to find the shortest sequence of (n, d) such that :math:`abs(\sum_i n_i \pi / 2 ^ {d_i} - angle) < tol`
    This is to find a sequence of rotations for a given angle.

    Parameters
    ----------
    angle : float
        The angle to approximate
    tol : float
        Tolerance to use
    """
    angle %= 2 * np.pi
    rest = angle / np.pi

    # Max value of `n`
    n_max = 2**IMMEDIATE_BITS - 1

    nds = []
    while rest > tol:
        # Find the largest `d` such that `rest <= n_max / 2 ^ d`
        d = int(np.floor(np.log2(n_max / rest)))
        # Find largest `n` such that `rest >= n / 2 ^ d`
        n = int(np.floor(rest * 2**d))
        # Shouldn't happen, but lets make sure
        assert n <= n_max, "Something went wrong, n is bigger than n_max"
        nds.append((n, d))
        rest -= n / 2**d

    # Check if some of the (n, d)'s can be simplified, i.e. if `n = b * 2 ^ m` for some `m` and `b`
    for i, (n, d) in enumerate(nds):
        n_new, d_new = n, d
        while (n_new % 2) == 0:
            n_new, d_new = (int(n_new / 2), d_new - 1)
        nds[i] = (n_new, d_new)
    nds = [(n, d) for (n, d) in nds if d < 32]
    return nds



from netqasm.sdk import EPRSocket
from netqasm.sdk.external import NetQASMConnection, Socket


def main(app_config=None):
    epr_socket_alice = EPRSocket("alice")

    socket_bob = Socket("repeater", "bob", log_config=app_config.log_config)
    epr_socket_bob = EPRSocket("bob")

    repeater = NetQASMConnection(
        app_name=app_config.app_name,
        log_config=app_config.log_config,
        epr_sockets=[epr_socket_alice, epr_socket_bob],
    )

    with repeater:
        # Wait for entanglement with Alice
        epr_alice = epr_socket_alice.recv_keep()[0]

        # Create entanglement with Bob
        epr_bob = epr_socket_bob.create_keep()[0]

        # Teleport qubit that is entangled with Alice to Bob
        epr_alice.cnot(epr_bob)
        epr_alice.H()
        m1 = epr_alice.measure()
        m2 = epr_bob.measure()

    m1, m2 = int(m1), int(m2)

    # Send teleportation corrections to Bob
    msg = str((m1, m2))
    socket_bob.send(msg)

    return {"m1": m1, "m2": m2}


if __name__ == "__main__":
    main()



import numpy as np
import pytest

from netqasm.logging.glob import get_netqasm_logger
from netqasm.runtime.application import default_app_instance
from netqasm.runtime.settings import Simulator, get_simulator
from netqasm.sdk import EPRSocket, Qubit
from netqasm.sdk.external import NetQASMConnection, Socket, simulate_application

logger = get_netqasm_logger()


def run_alice():
    socket = Socket("Alice", "Bob")
    epr_socket = EPRSocket("Bob")
    with NetQASMConnection("Alice", epr_sockets=[epr_socket]) as alice:
        # Create a qubit
        q = Qubit(alice)
        q.H()

        # Create entanglement
        epr = epr_socket.create()[0]

        # Teleport
        q.cnot(epr)
        q.H()
        m1 = q.measure()
        m2 = epr.measure()

    logger.info(f"m1, m2 = {m1}, {m2}")

    # Send the correction information
    msg = str((int(m1), int(m2)))
    socket.send(msg)


def run_bob():
    socket = Socket("Bob", "Alice")
    epr_socket = EPRSocket("Alice")
    with NetQASMConnection("Bob", epr_sockets=[epr_socket]) as bob:
        epr = epr_socket.recv()[0]
        bob.flush()

        # Get the corrections
        msg = socket.recv()
        logger.info(f"Bob got corrections: {msg}")
        m1, m2 = eval(msg)
        if m2 == 1:
            epr.X()
        if m1 == 1:
            epr.Z()


def post_function(backend):
    state = backend.nodes["Bob"].qmemory._get_qubits(0)[0].qstate.qrepr.reduced_dm()
    logger.info(f"state = {state}")
    expected = np.array([[0.5, 0.5], [0.5, 0.5]])
    logger.info(f"expected = {expected}")
    assert np.all(np.isclose(expected, state))


@pytest.mark.skipif(
    get_simulator() == Simulator.SIMULAQRON,
    reason="SimulaQron does not yet support a post_function",
)
def test_teleport():
    app_instance = default_app_instance(
        [
            ("Alice", run_alice),
            ("Bob", run_bob),
        ]
    )
    simulate_application(
        app_instance,
        use_app_config=False,
        post_function=post_function,
        enable_logging=False,
    )



"""
Network stack interface for simulators.

This module provides the `BaseNetworkStack` class which can be used by simulators
as a base class for modeling the network stack.
"""

import abc
from dataclasses import dataclass
from typing import Any, Generator, Union

from netqasm.qlink_compat import (
    LinkLayerCreate,
    LinkLayerOKTypeK,
    LinkLayerOKTypeM,
    LinkLayerRecv,
)

# Number of elements in a create request etc
# NOTE minus 2 since remote_node_id and epr_socket_id comes as immediates
CREATE_FIELDS = len(LinkLayerCreate._fields) - 2
OK_FIELDS_K = len(LinkLayerOKTypeK._fields)
OK_FIELDS_M = len(LinkLayerOKTypeM._fields)


@dataclass(eq=True, frozen=True)
class Address:
    node_id: int
    epr_socket_id: int


class BaseNetworkStack(abc.ABC):
    """Base class for a (extremly simplified) network stack in simulators.

    This class can be used as a simplified network stack that can handle requests
    that adhere to the QLink-Interface.
    """

    @abc.abstractmethod
    def put(self, request: Union[LinkLayerCreate, LinkLayerRecv]) -> None:
        """Put a link layer request to the network stack"""
        pass

    @abc.abstractmethod
    def setup_epr_socket(
        self,
        epr_socket_id: int,
        remote_node_id: int,
        remote_epr_socket_id: int,
        timeout: float = 1.0,
    ) -> Generator[Any, None, None]:
        """Ask the network stack to open an EPR socket."""
        pass

    @abc.abstractmethod
    def get_purpose_id(self, remote_node_id: int, epr_socket_id: int) -> int:
        pass



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import List, Optional, Union

from netqasm.lang import encoding
from netqasm.lang.operand import (
    Address,
    ArrayEntry,
    ArraySlice,
    Immediate,
    Operand,
    Register,
)
from netqasm.util.log import HostLine
from netqasm.util.string import rspaces

# Abstract base instruction types. Should not be instantiated directly.


@dataclass  # type: ignore
class NetQASMInstruction(ABC):
    """
    Base NetQASM instruction class.
    """

    id: int = -1
    mnemonic: str = ""
    lineno: Optional[HostLine] = None

    @property
    @abstractmethod
    def operands(self) -> List[Operand]:
        pass

    @classmethod
    @abstractmethod
    def deserialize_from(cls, raw: bytes) -> "NetQASMInstruction":
        pass

    @abstractmethod
    def serialize(self) -> bytes:
        pass

    @classmethod
    @abstractmethod
    def from_operands(cls, operands: List[Union[Operand, int]]) -> "NetQASMInstruction":
        pass

    def writes_to(self) -> List[Register]:
        """Returns a list of Registers that this instruction writes to"""
        return []

    def __str__(self):
        return self._build_str(show_lineno=False)

    @property
    def debug_str(self):
        return self._build_str(show_lineno=True)

    def _get_lineno_str(self):
        if self.lineno is None:
            lineno = "()"
        else:
            lineno = f"({self.lineno})"
        return f"{rspaces(lineno, min_chars=5)} "

    def _build_str(self, show_lineno=False):
        if show_lineno:
            lineno_str = self._get_lineno_str()
        else:
            lineno_str = ""
        return f"{lineno_str}{self._pretty_print()}"

    def _pretty_print(self):
        return f"{self.__class__}"


@dataclass
class NoOperandInstruction(NetQASMInstruction):
    """
    An instruction with no operands.
    """

    @property
    def operands(self) -> List[Operand]:
        return []

    @classmethod
    def deserialize_from(cls, raw: bytes):
        c_struct = encoding.NoOperandCommand.from_buffer_copy(raw)
        assert c_struct.id == cls.id
        return cls(id=cls.id)

    def serialize(self) -> bytes:
        c_struct = encoding.NoOperandCommand(id=self.id)
        return bytes(c_struct)

    @classmethod
    def from_operands(cls, operands: List[Union[Operand, int]]):
        assert len(operands) == 0
        return cls(id=cls.id)

    def _pretty_print(self):
        return f"{self.mnemonic}"


@dataclass
class RegInstruction(NetQASMInstruction):
    """
    An instruction with 1 Register operand.
    """

    reg: Register = None  # type: ignore

    @property
    def operands(self) -> List[Operand]:
        return [self.reg]

    @classmethod
    def deserialize_from(cls, raw: bytes):
        c_struct = encoding.RegCommand.from_buffer_copy(raw)
        assert c_struct.id == cls.id
        reg = Register.from_raw(c_struct.reg)
        return cls(id=cls.id, reg=reg)

    def serialize(self) -> bytes:
        c_struct = encoding.RegCommand(id=self.id, reg=self.reg.cstruct)
        return bytes(c_struct)

    @classmethod
    def from_operands(cls, operands: List[Union[Operand, int]]):
        assert len(operands) == 1
        reg = operands[0]
        assert isinstance(reg, Register)
        return cls(id=cls.id, reg=reg)

    def _pretty_print(self):
        return f"{self.mnemonic} {str(self.reg)}"


@dataclass
class RegRegInstruction(NetQASMInstruction):
    """
    An instruction with 2 Register operands.
    """

    reg0: Register = None  # type: ignore
    reg1: Register = None  # type: ignore

    @property
    def operands(self) -> List[Operand]:
        return [self.reg0, self.reg1]

    @classmethod
    def deserialize_from(cls, raw: bytes):
        c_struct = encoding.RegRegCommand.from_buffer_copy(raw)
        assert c_struct.id == cls.id
        reg0 = Register.from_raw(c_struct.reg0)
        reg1 = Register.from_raw(c_struct.reg1)
        return cls(reg0=reg0, reg1=reg1)

    def serialize(self) -> bytes:
        c_struct = encoding.RegRegCommand(
            id=self.id, reg0=self.reg0.cstruct, reg1=self.reg1.cstruct
        )
        return bytes(c_struct)

    @classmethod
    def from_operands(cls, operands: List[Union[Operand, int]]):
        assert len(operands) == 2
        reg0, reg1 = operands
        assert isinstance(reg0, Register)
        assert isinstance(reg1, Register)
        return cls(reg0=reg0, reg1=reg1)

    def _pretty_print(self):
        return f"{self.mnemonic} {str(self.reg0)} {str(self.reg1)}"


@dataclass
class RegImmImmInstruction(NetQASMInstruction):
    """
    An instruction with 1 Register operand followed by 2 Immediate operands.
    """

    reg: Register = None  # type: ignore
    imm0: Immediate = None  # type: ignore
    imm1: Immediate = None  # type: ignore

    @property
    def operands(self) -> List[Operand]:
        return [self.reg, self.imm0, self.imm1]

    @classmethod
    def deserialize_from(cls, raw: bytes):
        c_struct = encoding.RegImmImmCommand.from_buffer_copy(raw)
        assert c_struct.id == cls.id
        reg = Register.from_raw(c_struct.reg)
        imm0 = Immediate(value=c_struct.imm0)
        imm1 = Immediate(value=c_struct.imm1)
        return cls(reg=reg, imm0=imm0, imm1=imm1)

    def serialize(self) -> bytes:
        c_struct = encoding.RegImmImmCommand(
            id=self.id, reg=self.reg.cstruct, imm0=self.imm0.value, imm1=self.imm1.value
        )
        return bytes(c_struct)

    @classmethod
    def from_operands(cls, operands: List[Union[Operand, int]]):
        assert len(operands) == 3
        reg = operands[0]
        assert isinstance(reg, Register)

        imm0 = (
            Immediate(value=operands[1])
            if isinstance(operands[1], int)
            else operands[1]
        )
        assert isinstance(imm0, Immediate)

        imm1 = (
            Immediate(value=operands[2])
            if isinstance(operands[2], int)
            else operands[2]
        )
        assert isinstance(imm1, Immediate)

        return cls(reg=reg, imm0=imm0, imm1=imm1)

    def _pretty_print(self):
        return f"{self.mnemonic} {str(self.reg)} {str(self.imm0)} {str(self.imm1)}"


@dataclass
class RegRegImmImmInstruction(NetQASMInstruction):
    """
    An instruction with 2 Register operands followed by 2 Immediate operands.
    """

    reg0: Register = None  # type: ignore
    reg1: Register = None  # type: ignore
    imm0: Immediate = None  # type: ignore
    imm1: Immediate = None  # type: ignore

    @property
    def operands(self) -> List[Operand]:
        return [self.reg0, self.reg1, self.imm0, self.imm1]

    @classmethod
    def deserialize_from(cls, raw: bytes):
        c_struct = encoding.RegRegImmImmCommand.from_buffer_copy(raw)
        assert c_struct.id == cls.id
        reg0 = Register.from_raw(c_struct.reg0)
        reg1 = Register.from_raw(c_struct.reg1)
        imm0 = Immediate(value=c_struct.imm0)
        imm1 = Immediate(value=c_struct.imm1)
        return cls(reg0=reg0, reg1=reg1, imm0=imm0, imm1=imm1)

    def serialize(self) -> bytes:
        c_struct = encoding.RegRegImmImmCommand(
            id=self.id,
            reg0=self.reg0.cstruct,
            reg1=self.reg1.cstruct,
            imm0=self.imm0.value,
            imm1=self.imm1.value,
        )
        return bytes(c_struct)

    @classmethod
    def from_operands(cls, operands: List[Union[Operand, int]]):
        assert len(operands) == 4
        reg0, reg1, imm0, imm1 = operands
        assert isinstance(reg0, Register)
        assert isinstance(reg1, Register)
        assert isinstance(imm0, int) or isinstance(imm0, Immediate)
        if isinstance(imm0, int):
            imm0 = Immediate(value=imm0)
        assert isinstance(imm1, int) or isinstance(imm1, Immediate)
        if isinstance(imm1, int):
            imm1 = Immediate(value=imm1)
        return cls(reg0=reg0, reg1=reg1, imm0=imm0, imm1=imm1)

    def _pretty_print(self):
        return f"{self.mnemonic} {str(self.reg0)} {str(self.reg1)} {str(self.imm0)} {str(self.imm1)}"


@dataclass
class RegRegImm4Instruction(NetQASMInstruction):
    """
    An instruction with 2 Register operands followed by 4 Immediate operands.
    """

    reg0: Register = None  # type: ignore
    reg1: Register = None  # type: ignore
    imm0: Immediate = None  # type: ignore
    imm1: Immediate = None  # type: ignore
    imm2: Immediate = None  # type: ignore
    imm3: Immediate = None  # type: ignore

    @property
    def operands(self) -> List[Operand]:
        return [self.reg0, self.reg1, self.imm0, self.imm1, self.imm2, self.imm3]

    @classmethod
    def deserialize_from(cls, raw: bytes):
        c_struct = encoding.RegRegImm4Command.from_buffer_copy(raw)
        assert c_struct.id == cls.id
        reg0 = Register.from_raw(c_struct.reg0)
        reg1 = Register.from_raw(c_struct.reg1)
        imm0 = Immediate(value=c_struct.imm0)
        imm1 = Immediate(value=c_struct.imm1)
        imm2 = Immediate(value=c_struct.imm2)
        imm3 = Immediate(value=c_struct.imm3)
        return cls(reg0=reg0, reg1=reg1, imm0=imm0, imm1=imm1, imm2=imm2, imm3=imm3)

    def serialize(self) -> bytes:
        c_struct = encoding.RegRegImm4Command(
            id=self.id,
            reg0=self.reg0.cstruct,
            reg1=self.reg1.cstruct,
            imm0=self.imm0.value,
            imm1=self.imm1.value,
            imm2=self.imm2.value,
            imm3=self.imm3.value,
        )
        return bytes(c_struct)

    @classmethod
    def from_operands(cls, operands: List[Union[Operand, int]]):
        assert len(operands) == 6
        reg0, reg1, imm0, imm1, imm2, imm3 = operands
        assert isinstance(reg0, Register)
        assert isinstance(reg1, Register)
        assert isinstance(imm0, int) or isinstance(imm0, Immediate)
        if isinstance(imm0, int):
            imm0 = Immediate(value=imm0)
        assert isinstance(imm1, int) or isinstance(imm1, Immediate)
        if isinstance(imm1, int):
            imm1 = Immediate(value=imm1)
        assert isinstance(imm2, int) or isinstance(imm2, Immediate)
        if isinstance(imm2, int):
            imm2 = Immediate(value=imm2)
        assert isinstance(imm3, int) or isinstance(imm3, Immediate)
        if isinstance(imm3, int):
            imm3 = Immediate(value=imm3)
        return cls(reg0=reg0, reg1=reg1, imm0=imm0, imm1=imm1, imm2=imm2, imm3=imm3)

    def _pretty_print(self):
        return (
            f"{self.mnemonic} {str(self.reg0)} {str(self.reg1)} {str(self.imm0)} "
            f"{str(self.imm1)} {str(self.imm2)} {str(self.imm3)}"
        )


@dataclass
class RegRegRegInstruction(NetQASMInstruction):
    """
    An instruction with 3 Register operands.
    """

    reg0: Register = None  # type: ignore
    reg1: Register = None  # type: ignore
    reg2: Register = None  # type: ignore

    @property
    def operands(self) -> List[Operand]:
        return [self.reg0, self.reg1, self.reg2]

    @classmethod
    def deserialize_from(cls, raw: bytes):
        c_struct = encoding.RegRegRegCommand.from_buffer_copy(raw)
        assert c_struct.id == cls.id
        reg0 = Register.from_raw(c_struct.reg0)
        reg1 = Register.from_raw(c_struct.reg1)
        reg2 = Register.from_raw(c_struct.reg2)
        return cls(reg0=reg0, reg1=reg1, reg2=reg2)

    def serialize(self) -> bytes:
        c_struct = encoding.RegRegRegCommand(
            id=self.id,
            reg0=self.reg0.cstruct,
            reg1=self.reg1.cstruct,
            reg2=self.reg2.cstruct,
        )
        return bytes(c_struct)

    @classmethod
    def from_operands(cls, operands: List[Union[Operand, int]]):
        assert len(operands) == 3
        reg0, reg1, reg2 = operands
        assert isinstance(reg0, Register)
        assert isinstance(reg1, Register)
        assert isinstance(reg2, Register)
        return cls(reg0=reg0, reg1=reg1, reg2=reg2)

    def _pretty_print(self):
        return f"{self.mnemonic} {str(self.reg0)} {str(self.reg1)} {str(self.reg2)}"


@dataclass
class RegRegRegRegInstruction(NetQASMInstruction):
    """
    An instruction with 4 Register operands.
    """

    reg0: Register = None  # type: ignore
    reg1: Register = None  # type: ignore
    reg2: Register = None  # type: ignore
    reg3: Register = None  # type: ignore

    @property
    def operands(self) -> List[Operand]:
        return [self.reg0, self.reg1, self.reg2, self.reg3]

    @classmethod
    def deserialize_from(cls, raw: bytes):
        c_struct = encoding.RegRegRegRegCommand.from_buffer_copy(raw)
        assert c_struct.id == cls.id
        reg0 = Register.from_raw(c_struct.reg0)
        reg1 = Register.from_raw(c_struct.reg1)
        reg2 = Register.from_raw(c_struct.reg2)
        reg3 = Register.from_raw(c_struct.reg3)
        return cls(reg0=reg0, reg1=reg1, reg2=reg2, reg3=reg3)

    def serialize(self) -> bytes:
        c_struct = encoding.RegRegRegRegCommand(
            id=self.id,
            reg0=self.reg0.cstruct,
            reg1=self.reg1.cstruct,
            reg2=self.reg2.cstruct,
            reg3=self.reg3.cstruct,
        )
        return bytes(c_struct)

    @classmethod
    def from_operands(cls, operands: List[Union[Operand, int]]):
        assert len(operands) == 4
        reg0, reg1, reg2, reg3 = operands
        assert isinstance(reg0, Register)
        assert isinstance(reg1, Register)
        assert isinstance(reg2, Register)
        assert isinstance(reg3, Register)
        return cls(
            reg0=reg0,
            reg1=reg1,
            reg2=reg2,
            reg3=reg3,
        )

    def _pretty_print(self):
        return f"{self.mnemonic} {str(self.reg0)} {str(self.reg1)} {str(self.reg2)} {str(self.reg3)}"


@dataclass
class ImmInstruction(NetQASMInstruction):
    """
    An instruction with 1 Immediate operand.
    """

    imm: Immediate = None  # type: ignore

    @property
    def operands(self) -> List[Operand]:
        return [self.imm]

    @classmethod
    def deserialize_from(cls, raw: bytes):
        c_struct = encoding.ImmCommand.from_buffer_copy(raw)
        assert c_struct.id == cls.id
        imm = Immediate(value=c_struct.imm)
        return cls(imm=imm)

    def serialize(self) -> bytes:
        c_struct = encoding.ImmCommand(id=self.id, imm=self.imm.value)
        return bytes(c_struct)

    @classmethod
    def from_operands(cls, operands: List[Union[Operand, int]]):
        assert len(operands) == 1
        imm = operands[0]
        assert isinstance(imm, int) or isinstance(imm, Immediate)
        if isinstance(imm, int):
            imm = Immediate(value=imm)

        return cls(imm=imm)

    def _pretty_print(self):
        return f"{self.mnemonic} {str(self.imm)}"


@dataclass
class ImmImmInstruction(NetQASMInstruction):
    """
    An instruction with 2 Immediate operands.
    """

    imm0: Immediate = None  # type: ignore
    imm1: Immediate = None  # type: ignore

    @property
    def operands(self) -> List[Operand]:
        return [self.imm0, self.imm1]

    @classmethod
    def deserialize_from(cls, raw: bytes):
        c_struct = encoding.ImmImmCommand.from_buffer_copy(raw)
        assert c_struct.id == cls.id
        imm0 = Immediate(value=c_struct.imm0)
        imm1 = Immediate(value=c_struct.imm1)
        return cls(imm0=imm0, imm1=imm1)

    def serialize(self) -> bytes:
        c_struct = encoding.ImmImmCommand(
            id=self.id, imm0=self.imm0.value, imm1=self.imm1.value
        )
        return bytes(c_struct)

    @classmethod
    def from_operands(cls, operands: List[Union[Operand, int]]):
        assert len(operands) == 2
        imm0 = operands[0]
        imm1 = operands[1]
        assert isinstance(imm0, int) or isinstance(imm0, Immediate)
        if isinstance(imm0, int):
            imm0 = Immediate(value=imm0)
        assert isinstance(imm1, int) or isinstance(imm1, Immediate)
        if isinstance(imm1, int):
            imm1 = Immediate(value=imm1)
        return cls(imm0=imm0, imm1=imm1)

    def _pretty_print(self):
        return f"{self.mnemonic} {str(self.imm0)} {str(self.imm1)}"


@dataclass
class RegRegImmInstruction(NetQASMInstruction):
    """
    An instruction with 2 Register operands and one Immediate operand.
    """

    reg0: Register = None  # type: ignore
    reg1: Register = None  # type: ignore
    imm: Immediate = None  # type: ignore

    @property
    def operands(self) -> List[Operand]:
        return [self.reg0, self.reg1, self.imm]

    @classmethod
    def deserialize_from(cls, raw: bytes):
        c_struct = encoding.RegRegImmCommand.from_buffer_copy(raw)
        assert c_struct.id == cls.id
        reg0 = Register.from_raw(c_struct.reg0)
        reg1 = Register.from_raw(c_struct.reg1)
        imm = Immediate(value=c_struct.imm)
        return cls(reg0=reg0, reg1=reg1, imm=imm)

    def serialize(self) -> bytes:
        c_struct = encoding.RegRegImmCommand(
            id=self.id,
            reg0=self.reg0.cstruct,
            reg1=self.reg1.cstruct,
            imm=self.imm.value,
        )
        return bytes(c_struct)

    @classmethod
    def from_operands(cls, operands: List[Union[Operand, int]]):
        assert len(operands) == 3
        reg0, reg1, imm = operands
        assert isinstance(reg0, Register)
        assert isinstance(reg1, Register)
        assert isinstance(imm, int) or isinstance(imm, Immediate)
        if isinstance(imm, int):
            imm = Immediate(value=imm)
        return cls(reg0=reg0, reg1=reg1, imm=imm)

    def _pretty_print(self):
        return f"{self.mnemonic} {str(self.reg0)} {str(self.reg1)} {str(self.imm)}"


@dataclass
class RegImmInstruction(NetQASMInstruction):
    """
    An instruction with 1 Register operand and one Immediate operand.
    """

    reg: Register = None  # type: ignore
    imm: Immediate = None  # type: ignore

    @property
    def operands(self) -> List[Operand]:
        return [self.reg, self.imm]

    @classmethod
    def deserialize_from(cls, raw: bytes):
        c_struct = encoding.RegImmCommand.from_buffer_copy(raw)
        assert c_struct.id == cls.id
        reg = Register.from_raw(c_struct.reg)
        imm = Immediate(value=c_struct.imm)
        return cls(reg=reg, imm=imm)

    def serialize(self) -> bytes:
        c_struct = encoding.RegImmCommand(
            id=self.id, reg=self.reg.cstruct, imm=self.imm.value
        )
        return bytes(c_struct)

    @classmethod
    def from_operands(cls, operands: List[Union[Operand, int]]):
        assert len(operands) == 2
        reg, imm = operands
        assert isinstance(reg, Register)
        assert isinstance(imm, int) or isinstance(imm, Immediate)
        if isinstance(imm, int):
            imm = Immediate(value=imm)
        return cls(reg=reg, imm=imm)

    def _pretty_print(self):
        return f"{self.mnemonic} {str(self.reg)} {str(self.imm)}"


@dataclass
class RegEntryInstruction(NetQASMInstruction):
    """
    An instruction with 1 Register operand and one ArrayEntry operand.
    """

    reg: Register = None  # type: ignore
    entry: ArrayEntry = None  # type: ignore

    @property
    def operands(self) -> List[Operand]:
        return [self.reg, self.entry]

    @classmethod
    def deserialize_from(cls, raw: bytes):
        c_struct = encoding.RegEntryCommand.from_buffer_copy(raw)
        assert c_struct.id == cls.id
        reg = Register.from_raw(c_struct.reg)
        entry = ArrayEntry.from_raw(c_struct.entry)
        return cls(reg=reg, entry=entry)

    def serialize(self) -> bytes:
        c_struct = encoding.RegEntryCommand(
            id=self.id, reg=self.reg.cstruct, entry=self.entry.cstruct
        )
        return bytes(c_struct)

    @classmethod
    def from_operands(cls, operands: List[Union[Operand, int]]):
        assert len(operands) == 2
        reg, entry = operands
        assert isinstance(reg, Register)
        assert isinstance(entry, ArrayEntry)
        return cls(reg=reg, entry=entry)

    def _pretty_print(self):
        return f"{self.mnemonic} {str(self.reg)} {str(self.entry)}"


@dataclass
class RegAddrInstruction(NetQASMInstruction):
    """
    An instruction with 1 Register operand and one Address operand.
    """

    reg: Register = None  # type: ignore
    address: Address = None  # type: ignore

    @property
    def operands(self) -> List[Operand]:
        return [self.reg, self.address]

    @classmethod
    def deserialize_from(cls, raw: bytes):
        c_struct = encoding.RegAddrCommand.from_buffer_copy(raw)
        assert c_struct.id == cls.id
        reg = Register.from_raw(c_struct.reg)
        address = Address.from_raw(c_struct.addr)
        return cls(reg=reg, address=address)

    def serialize(self) -> bytes:
        c_struct = encoding.RegAddrCommand(
            id=self.id, reg=self.reg.cstruct, addr=self.address.cstruct
        )
        return bytes(c_struct)

    @classmethod
    def from_operands(cls, operands: List[Union[Operand, int]]):
        assert len(operands) == 2
        reg, addr = operands
        assert isinstance(reg, Register)
        assert isinstance(addr, Address)
        return cls(reg=reg, address=addr)

    def _pretty_print(self):
        return f"{self.mnemonic} {str(self.reg)} {str(self.address)}"


@dataclass
class ArrayEntryInstruction(NetQASMInstruction):
    """
    An instruction with 1 ArrayEntry operand and one Address operand.
    """

    entry: ArrayEntry = None  # type: ignore

    @property
    def operands(self) -> List[Operand]:
        return [self.entry]

    @classmethod
    def deserialize_from(cls, raw: bytes):
        c_struct = encoding.ArrayEntryCommand.from_buffer_copy(raw)
        assert c_struct.id == cls.id
        entry = ArrayEntry.from_raw(c_struct.entry)
        return cls(entry=entry)

    def serialize(self) -> bytes:
        c_struct = encoding.ArrayEntryCommand(id=self.id, entry=self.entry.cstruct)
        return bytes(c_struct)

    @classmethod
    def from_operands(cls, operands: List[Union[Operand, int]]):
        assert len(operands) == 1
        entry = operands[0]
        assert isinstance(entry, ArrayEntry)
        return cls(entry=entry)

    def _pretty_print(self):
        return f"{self.mnemonic} {str(self.entry)}"


@dataclass
class ArraySliceInstruction(NetQASMInstruction):
    """
    An instruction with 1 ArraySlice operand.
    """

    slice: ArraySlice = None  # type: ignore

    @property
    def operands(self) -> List[Operand]:
        return [self.slice]

    @classmethod
    def deserialize_from(cls, raw: bytes):
        c_struct = encoding.ArraySliceCommand.from_buffer_copy(raw)
        assert c_struct.id == cls.id
        slice = ArraySlice.from_raw(c_struct.slice)
        return cls(slice=slice)

    def serialize(self) -> bytes:
        c_struct = encoding.ArraySliceCommand(id=self.id, slice=self.slice.cstruct)
        return bytes(c_struct)

    @classmethod
    def from_operands(cls, operands: List[Union[Operand, int]]):
        assert len(operands) == 1
        slice = operands[0]
        assert isinstance(slice, ArraySlice)
        return cls(slice=slice)

    def _pretty_print(self):
        return f"{self.mnemonic} {str(self.slice)}"


@dataclass
class AddrInstruction(NetQASMInstruction):
    """
    An instruction with 1 Address operand.
    """

    address: Address = None  # type: ignore

    @property
    def operands(self) -> List[Operand]:
        return [self.address]

    @classmethod
    def deserialize_from(cls, raw: bytes):
        c_struct = encoding.AddrCommand.from_buffer_copy(raw)
        assert c_struct.id == cls.id
        addr = Address.from_raw(c_struct.addr)
        return cls(address=addr)

    def serialize(self) -> bytes:
        c_struct = encoding.AddrCommand(id=self.id, addr=self.address.cstruct)
        return bytes(c_struct)

    @classmethod
    def from_operands(cls, operands: List[Union[Operand, int]]):
        assert len(operands) == 1
        addr = operands[0]
        assert isinstance(addr, Address)
        return cls(address=addr)

    def _pretty_print(self):
        return f"{self.mnemonic} {str(self.address)}"


@dataclass
class Reg5Instruction(NetQASMInstruction):
    """
    An instruction with 5 Register operands.
    """

    reg0: Register = None  # type: ignore
    reg1: Register = None  # type: ignore
    reg2: Register = None  # type: ignore
    reg3: Register = None  # type: ignore
    reg4: Register = None  # type: ignore

    @property
    def operands(self) -> List[Operand]:
        return [self.reg0, self.reg1, self.reg2, self.reg3, self.reg4]

    @classmethod
    def deserialize_from(cls, raw: bytes):
        c_struct = encoding.Reg5Command.from_buffer_copy(raw)
        assert c_struct.id == cls.id
        return cls(
            reg0=Register.from_raw(c_struct.reg0),
            reg1=Register.from_raw(c_struct.reg1),
            reg2=Register.from_raw(c_struct.reg2),
            reg3=Register.from_raw(c_struct.reg3),
            reg4=Register.from_raw(c_struct.reg4),
        )

    def serialize(self) -> bytes:
        c_struct = encoding.Reg5Command(
            id=self.id,
            reg0=self.reg0.cstruct,
            reg1=self.reg1.cstruct,
            reg2=self.reg2.cstruct,
            reg3=self.reg3.cstruct,
            reg4=self.reg4.cstruct,
        )
        return bytes(c_struct)

    @classmethod
    def from_operands(cls, operands: List[Union[Operand, int]]):
        assert len(operands) == 5
        reg0, reg1, reg2, reg3, reg4 = operands
        assert isinstance(reg0, Register)
        assert isinstance(reg1, Register)
        assert isinstance(reg2, Register)
        assert isinstance(reg3, Register)
        assert isinstance(reg4, Register)
        return cls(
            reg0=reg0,
            reg1=reg1,
            reg2=reg2,
            reg3=reg3,
            reg4=reg4,
        )

    def _pretty_print(self):
        return f"{self.mnemonic} {str(self.reg0)} {str(self.reg1)} {str(self.reg2)} \
{str(self.reg3)} {str(self.reg4)}"


@dataclass
class DebugInstruction(NetQASMInstruction):
    text: str = ""

    @property
    def operands(self) -> List[Operand]:
        pass

    @classmethod
    def deserialize_from(cls, raw: bytes):
        pass

    def serialize(self) -> bytes:
        return b""

    @classmethod
    def from_operands(cls, operands: List[Union[Operand, int]]):
        pass

    def _pretty_print(self):
        return f"# {self.text}"
